# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2020 by Hildo Guillardi Junior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
import os
import re
from collections import OrderedDict
from .. import DEBUG_OVERVIEW, DEBUG_OBSESSIVE, SEPRTR, W_FLDOVR
from ..distributors import get_distributors_iter

__all__ = ['eda_class', 'field_name_translations']


# Generate a dictionary to translate all the different ways people might want
# to refer to part numbers, vendor numbers, manufacture name and such.
field_name_translations = {
    # footprint
    'package': 'footprint',
    'pcb footprint': 'footprint',
    'pcb package': 'footprint',  # Used at Proteus.
    # manf#
    'mpn': 'manf#',
    'pn': 'manf#',
    'manf_num': 'manf#',
    'manf-num': 'manf#',
    'mfg_num': 'manf#',
    'mfg-num': 'manf#',
    'mfg#': 'manf#',
    'mfg part#': 'manf#',
    'mfr. no': 'manf#',
    'man#': 'manf#',
    'man_num': 'manf#',
    'man-num': 'manf#',
    'manpartno': 'manf#',
    'manufacturer part number': 'manf#',  # Use on `http://upverter.com/`.
    'mnf_num': 'manf#',
    'mnf-num': 'manf#',
    'mnf#': 'manf#',
    'mfr_num': 'manf#',
    'mfr-num': 'manf#',
    'mfr#': 'manf#',
    'part-num': 'manf#',
    'part_num': 'manf#',
    'p#': 'manf#',
    'part#': 'manf#',
    'stock code': 'manf#',
    # manf
    'manufacturer': 'manf',
    'manufacturer name': 'manf',  # Used for some web site tools to part generator in Altium.
    'mnf': 'manf',
    'man': 'manf',
    'mfg': 'manf',
    'mfr': 'manf',
    # refs
    'designator': 'refs',
    'part reference': 'refs',
    'reference': 'refs',
    'reference designator': 'refs',
    'references': 'refs',
    'customer no': 'refs',
    'parts': 'refs',
    'part': 'refs',
    # qty
    'order qty': 'qty',
    'quantity': 'qty',
    # Others fields used by KiCost and that have to be standardized.
    'pdf': 'datasheet',
    'description': 'desc',
    'nopop': 'dnp',
    'version': 'variant',
}
# Create the fields translate for each distributor submodule.
for stub in ['part#', '#', 'p#', 'pn', 'vendor#', 'vp#', 'vpn', 'num']:
    for dist in get_distributors_iter():
        if stub != '#':
            field_name_translations[dist + stub] = dist + '#'
        field_name_translations[dist + '_' + stub] = dist + '#'
        field_name_translations[dist + '-' + stub] = dist + '#'


class eda_class(object):
    registered = {}
    logger = None

    @staticmethod
    def register(eda):
        ''' Register a new EDA class '''
        eda_class.registered[eda.name] = eda

    @staticmethod
    def get_part_groups(eda, in_file, ignore_fields, variant, distributors):
        '''Get groups of identical parts from a file and return them as a dictionary.
           @param in_file `str()` with the file name.
           @param ignore_fields `list()` fields do be ignored on the read action.
           @param variant `str()` in regular expression to match with the design version of the BOM.
           A tuple of two values is returned:
           @return `dict()` of the parts designed. The keys are the componentes references.
           @return `dict()` of project information.
        '''
        parts, prj_info = eda_class.registered[eda].get_part_groups(in_file, distributors)
        parts = eda_class.process_fields(parts, variant, ignore_fields, distributors)
        return eda_class.remove_dnp_parts(parts, variant), prj_info

    @staticmethod
    def process_kicost_field(f, v, new_fields, ignore_fields, distributors, name_ori, ref):
        f = f.strip()
        if SEPRTR in f:
            # DISTRIBUTOR:FIELD
            # Separate the distributor from the field name
            idx = f.index(SEPRTR)
            dist = f[:idx]
            # Is this a supported distributor?
            dist_l = dist.lower()
            if dist_l in distributors:
                # Use the lower case version
                dist = dist_l
            # Translate the field name
            f = f[idx+1:].lower()
            if f in ignore_fields:
                return
            f = field_name_translations.get(f, f)
            if f in ignore_fields:
                return
            # Join both again
            f = dist + SEPRTR + f
        else:
            # FIELD
            # No distributor in the name
            # Adapt it
            f = f.lower()
            if f in ignore_fields:
                return
            f = field_name_translations.get(f, f)
            if f in ignore_fields:
                return
            # Is distributor related?
            if f in ('cat#', 'pricing', 'link'):
                # Add it to the default local distributor
                eda_class.logger.log(DEBUG_OBSESSIVE, 'Assigning name "{}" to "Local" distributor'.format(f))
                f = 'Local:' + f
        # Note: here we allow a field to "clear" another definition (assigning an empty value)
        # This is because we assume that "kicost*" fields are defined on purpose, not just inherited from the library
        new_fields[f] = v.strip()
        eda_class.logger.log(DEBUG_OBSESSIVE, '{} Field {} -> {}={}'.format(ref, name_ori, f, v))

    @staticmethod
    def process_fields(parts, variant, ignore_fields, distributors):
        new_parts = OrderedDict()
        for ref, fields in parts.items():
            new_fields = OrderedDict()
            # Add the fields from lowest to highest priority.
            # We even add empty fields, so we can "clear" a field using a higher priority mechanism.
            # 1) All fields without SEPRTR (lowest priority)
            for f, v in fields.items():
                old_name = f
                if SEPRTR in f:
                    continue
                # Just translate it
                f = f.lower().strip()
                if f in ignore_fields:
                    continue
                f = field_name_translations.get(f, f)
                if f in ignore_fields:
                    continue
                # Trim extra spaces in the value
                v = v.strip()
                already_defined = f in new_fields
                if not v and already_defined:
                    # For regular fields we avoid one alias clearing another.
                    # Example: if manf# was defined as XXX and now we have mnp='' we avoid getting manf#=''
                    continue
                if already_defined and new_fields[f]:
                    eda_class.logger.warning(W_FLDOVR+'Warning: in {} overwriting {}={} with {}={}'.format(ref, f, new_fields[f], old_name, v))
                new_fields[f] = v
            # 2) kicost:FIELD
            for f, v in fields.items():
                if f.startswith('kicost' + SEPRTR):
                    eda_class.process_kicost_field(f[7:], v, new_fields, ignore_fields, distributors, f, ref)
            # 3) kicost.VARIANT:  (highest priority)
            for f, v in fields.items():
                if not (SEPRTR in f and f.startswith('kicost.')):
                    continue
                sep_pos = f.index(SEPRTR)
                var = f[7:sep_pos]
                if not re.match(variant, var, flags=re.IGNORECASE):
                    # Not for the current variant
                    continue
                name = f[sep_pos+1:]
                eda_class.logger.log(DEBUG_OBSESSIVE, 'Matched Variant ... ' + var + '.' + name)
                eda_class.process_kicost_field(name, v, new_fields, ignore_fields, distributors, f, ref)
            # TODO: What about the other cases?
            # * Has ':' but doesn't start with "kicost"
            new_parts[ref] = new_fields
        return new_parts

    @staticmethod
    def file_eda_match(file_name):
        '''@brief Verify with which EDA the file matches.

           Return the EDA name with the file matches or `None` if not founded.
           @param file_name File `str` name.
           @return Name of the module corresponding to read the file or `None`to not recognized.
        '''
        try:
            file_handle = open(file_name, 'r')
            content = file_handle.read()
        except UnicodeDecodeError:  # Python 3 assumes UTF-8, try with Latin1 if it failed
            file_handle.close()
            file_handle = open(file_name, 'r', encoding='ISO-8859-1')
            content = file_handle.read()
        file_handle.close()
        extension = os.path.splitext(file_name)[1]
        for name, cls in eda_class.registered.items():
            if cls.file_eda_match(content, extension):
                return name
        return None

    @staticmethod
    def remove_dnp_parts(components, variant):
        '''@brief Remove the DNP parts or not assigned to the current variant.

           Remove components that are assigned to a variant that is not the current variant,
           or which are "do not populate" (DNP). (Any component that does not have a variant
           is assigned the current variant so it will not be removed unless it is also DNP.)

           @param components Part components in a `list()` of `dict()`, format given by the EDA modules.
           @return `list()` of `dict()`.
        '''
        eda_class.logger.log(DEBUG_OVERVIEW, '# Removing do not populate parts...')
        accepted_components = OrderedDict()
        for ref, fields in components.items():
            # Remove DNPs.
            dnp = fields.get('dnp', '0')
            # Interpret empty strings as 0. See #471 discussion.
            if dnp == '':
                dnp = '0'
            try:
                dnp = float(dnp)
            except ValueError:
                # The field value must have been a string.
                # The docs says "any string" is DNP enabled.
                dnp = 1
            if dnp:
                continue
            # Get part variant. Prioritize local variants over global ones.
            variants = fields.get('variant', None)
            # Remove parts that are not assigned to the current variant.
            # If a part is not assigned to any variant, then it is never removed.
            if variants:
                # A part can be assigned to multiple variants. The part will not
                # be removed if any of its variants match the current variant.
                # Split the variants apart and abort the loop if any of them match.
                for v in re.split('[,;/ ]', variants):
                    if re.match(variant, v, flags=re.IGNORECASE):
                        break
                else:
                    # None of the variants matched, so skip/remove this part.
                    continue
            # The part was not removed, so add it to the list of accepted components.
            accepted_components[ref] = fields
        return accepted_components



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2022 by Salvador E. Tropea / Instituto Nacional de Tecnologia Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# I took the basic query from https://github.com/krzych82/api-client-python3.git

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft'
__company__ = 'Instituto Nacional de Tecnologia Industrial - Argentina'

# Libraries.
import sys
import pprint
import difflib
import collections
import json
import base64
import hmac
import hashlib
import time
if sys.version_info[0] < 3:
    from urllib import quote, urlencode
    from urllib2 import Request, urlopen, URLError
else:
    from urllib.parse import quote, urlencode
    from urllib.request import Request, urlopen
    from urllib.error import URLError

# KiCost definitions.
from .. import KiCostError, DistData, W_NOINFO, ERR_SCRAPE, W_APIFAIL
# Distributors definitions.
from .distributor import distributor_class, QueryCache, hide_secrets
from .log__ import debug_overview, debug_obsessive, debug_full, is_debug_full, warning

# Specs known by KiCost
SPEC_NAMES = {'tolerance': 'tolerance',
              'frequency': 'frequency',
              'power': 'power',
              'operating voltage': 'voltage',
              'manufacturer': 'manf',
              'temperature coefficient': 'temp_coeff'}

DIST_NAME = 'tme'
BASE_URL = 'https://api.tme.eu'
MAX_PARTS_PER_QUERY = 10

__all__ = ['api_tme']


def do_encode(string):
    if sys.version_info[0] < 3:
        return string
    return string.encode()


class TMEError(Exception):
    pass


class TME(object):
    def __init__(self, country, language, app_secret, token, cache, currency):
        # Adapt to what TME uses
        currency = currency.upper()
        country = country.upper()
        language = language.lower()
        # Store the options in the object
        self.currency = currency
        self.country = country
        self.language = language
        self.token = token
        self.app_secret = do_encode(app_secret)
        self.cache = cache
        # Check the language
        lng = self.get_languages()
        if language not in lng:
            self.language = 'EN'
            debug_overview('Language `{}` not supported using `EN`'.format(language))
        debug_overview('Using `{}`'.format(self.language))
        # Check the selected country
        cnt = self.get_countries()
        country_l = country.lower()
        cnt_data = next(iter(filter(lambda x: x['CountryId'].lower() == country_l, cnt)), None)
        if cnt_data is None:
            raise TMEError("Unsupported country `{}`".format(country))
        # Check if the currency is supported for this country
        if currency not in cnt_data['CurrencyList']:
            # Nope, use the default for this country
            self.currency = cnt_data['Currency']
            debug_overview('Currency `{}` not supported for `{}` using `{}`'.format(currency, country, self.currency))
        debug_overview('Using `{}` for `{}`'.format(self.currency, cnt_data['Name']))

    def _get_signature_base(self, url, params):
        params = collections.OrderedDict(sorted(params.items()))
        encoded_params = urlencode(params)
        signature_base = 'POST'+'&'+quote(url, '')+'&'+quote(encoded_params, '')
        return do_encode(signature_base)

    def _calculate_signature(self, url, params):
        hmac_value = hmac.new(self.app_secret, self._get_signature_base(url, params), hashlib.sha1).digest()
        return base64.encodebytes(hmac_value).rstrip()

    def request(self, endpoint, params, format='json'):
        url = BASE_URL+endpoint+'.'+format
        params['Token'] = self.token
        params['ApiSignature'] = self._calculate_signature(url, params)
        data = do_encode(urlencode(params))
        headers = {"Content-type": "application/x-www-form-urlencoded"}
        return Request(url, data, headers)

    def json_request(self, endpoint, params={}):
        reason = code = None
        debug_obsessive('TME json request endpoint {} params {}'.format(endpoint, params))
        try:
            response = urlopen(self.request(endpoint, params))
        except URLError as e:
            reason = e.reason
            code = e.code
        if code == 403:
            # Sometimes when we ask too fast
            time.sleep(1)
            try:
                response = urlopen(self.request(endpoint, params))
            except URLError as e:
                reason = e.reason
                code = e.code
        if reason:
            raise TMEError("Server error `{}` ({})".format(reason, code))
        data = json.loads(response.read())
        return data['Data']

    def get_countries(self):
        language = self.language
        data, loaded = self.cache.load_results('all', 'countries_'+language)
        if loaded:
            debug_obsessive('Data from cache: '+pprint.pformat(data))
            return data['CountryList']
        data = self.json_request('/Utils/GetCountries', {'Language': language})
        debug_obsessive('Data from web: '+pprint.pformat(data))
        self.cache.save_results('all', 'countries_'+language, data)
        return data['CountryList']

    def get_languages(self):
        data, loaded = self.cache.load_results('all', 'languages')
        if loaded:
            debug_obsessive('Data from cache: '+pprint.pformat(data))
            return data['LanguageList']
        data = self.json_request('/Utils/GetLanguages')
        debug_obsessive('Data from web: '+pprint.pformat(data))
        self.cache.save_results('all', 'languages', data)
        return data['LanguageList']

    def get_from_cache(self, symbol):
        full_name = symbol+'_'+self.country+'_'+self.language+'_'+self.currency
        data, loaded = self.cache.load_results('data', full_name)
        if loaded:
            debug_obsessive('Data from cache: '+pprint.pformat(data))
            return data
        return None

    def _get_products(self, symbols, kind, currency=False):
        parameters = {'Country': self.country,
                      'Language': self.language}
        if currency:
            parameters['Currency'] = self.currency
        for c, s in enumerate(symbols):
            parameters['SymbolList[{}]'.format(c)] = s
        debug_obsessive(parameters)
        data = self.json_request('/Products/'+kind, parameters)
        debug_obsessive('Data from web: '+pprint.pformat(data))
        return data['ProductList']

    def get_products(self, symbols):
        return self._get_products(symbols, 'GetProducts')

    def get_prices(self, symbols):
        return self._get_products(symbols, 'GetPricesAndStocks', currency=True)

    def get_parameters(self, symbols):
        return self._get_products(symbols, 'GetParameters')

    def get_files(self, symbols):
        return self._get_products(symbols, 'GetProductsFiles')

    def get_part_info(self, missing, known):
        products = self.get_products(missing)
        prices = self.get_prices(missing)
        parameters = self.get_parameters(missing)
        files = self.get_files(missing)
        # Join all the data
        all_data = {s: {} for s in missing}
        for d in products+prices+parameters+files:
            all_data[d['Symbol']].update(d)
        # Cache the data and it to the known
        qual_name = '_'+self.country+'_'+self.language+'_'+self.currency
        for k, v in all_data.items():
            self.cache.save_results('data', k+qual_name, v)
            known[k] = v

    def search(self, name):
        # Try to get the data from the cache
        full_name = name+'_'+self.country+'_'+self.language+'_'+self.currency
        data, loaded = self.cache.load_results('search', full_name)
        if loaded:
            debug_obsessive('Data from cache: '+pprint.pformat(data))
            return data
        # The docs suggest that spaces are allowed, but I get 400 (Bad Request)
        name = name.replace(' ', '-')
        parameters = {'Country': self.country,
                      'Language': self.language,
                      'SearchPlain': name,
                      'SearchOrder': 'PRICE_FIRST_QUANTITY',
                      'SearchOrderType': 'ASC'}
        data = self.json_request('/Products/Search', parameters)
        debug_obsessive('Data from web: '+pprint.pformat(data))
        total = data['Amount']
        if total > 20:
            # Get the rest of the matches
            product_list = data['ProductList']
            page = 2
            while len(product_list) < total:
                parameters = {'Country': self.country,
                              'Language': self.language,
                              'SearchPlain': name,
                              'SearchOrder': 'PRICE_FIRST_QUANTITY',
                              'SearchOrderType': 'ASC',
                              'SearchPage': str(page)}
                data = self.json_request('/Products/Search', parameters)
                product_list.extend(data['ProductList'])
                page = page+1
        else:
            product_list = data['ProductList']
        self.cache.save_results('search', full_name, product_list)
        return product_list


class api_tme(distributor_class):
    name = 'TME'
    type = 'api'
    # Currently enabled only by request
    enabled = True

    url = 'https://developers.tme.eu/en/'  # Web site API information.
    api_distributors = [DIST_NAME]
    # Options supported by this API
    config_options = {'token': str,
                      'app_secret': str,
                      'country': str,
                      'language': str}
    token = None
    app_secret = None
    country = 'US'
    language = 'EN'
    env_prefix = 'TME'
    env_ops = {}

    @staticmethod
    def configure(ops):
        cache_ttl = 7
        cache_path = None
        for k, v in ops.items():
            if k == 'token':
                api_tme.token = v
            elif k == 'app_secret':
                api_tme.app_secret = v
            elif k == 'country':
                api_tme.country = v
            elif k == 'language':
                api_tme.language = v
            elif k == 'enable':
                api_tme.enabled = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
        if api_tme.enabled and (api_tme.token is None or api_tme.app_secret is None):
            warning(W_APIFAIL, "Can't enable TME without a `token` and an `app_secret`")
            api_tme.enabled = False
        debug_obsessive('TME API configured to enabled {} token {} app_secret {} path {}'.
                        format(api_tme.enabled, hide_secrets(api_tme.token), hide_secrets(api_tme.app_secret), cache_path))
        if not api_tme.enabled:
            return
        # Configure the cache
        api_tme.cache = QueryCache(cache_path, cache_ttl)

    @staticmethod
    def _query_part_info(parts, distributors, currency):
        '''Fill-in the parts with price/qty/etc info from KitSpace.'''
        debug_overview('# Getting part data from TME ...')
        field_cat = DIST_NAME + '#'
        o = TME(api_tme.country, api_tme.language, api_tme.app_secret, api_tme.token, api_tme.cache, currency)
        #
        # First pass: collect the missing TME part numbers
        #
        # Setup progress bar to track progress of server queries.
        progress = distributor_class.progress(len(parts), distributor_class.logger)
        symbols = collections.OrderedDict()
        missing_symbols = []
        for part in parts:
            # Get the TME P/N for this part
            part_stock = part.fields.get(field_cat)
            if not part_stock:
                # We can't get information without a valid TME code (symbol)
                part_manf = part.fields.get('manf', '')
                part_code = part.fields.get('manf#')
                if part_code:
                    debug_obsessive('Searching P/N: {} from {}'.format(part_code, part_manf))
                    candidates = o.search(part_code)
                    debug_obsessive('Found {} matches'.format(len(candidates)))
                    part_stock = part.fields[field_cat] = _select_best(candidates, part_manf, part.qty_total_spreadsheet)
            if part_stock:
                # Add this symbol to the list of needed
                data = o.get_from_cache(part_stock)
                symbols[part_stock] = data
                if data is None:
                    missing_symbols.append(part_stock)
            progress.update(1)
        progress.close()
        #
        # Second pass: collect the missing data
        #
        if missing_symbols:
            n_unsolved = len(missing_symbols)
            progress = distributor_class.progress(n_unsolved, distributor_class.logger)
            for i in range(0, n_unsolved, MAX_PARTS_PER_QUERY):
                slc = slice(i, i+MAX_PARTS_PER_QUERY)
                o.get_part_info(missing_symbols[slc], symbols)
                progress.update(len(missing_symbols[slc]))
            progress.close()
        #
        # Fill the part information
        #
        for part in parts:
            part_stock = part.fields.get(field_cat)
            if not part_stock:
                warning(W_NOINFO, 'No information found at TME for part/s \'{}\''.format(part.refs))
                continue
            data = symbols[part_stock]
            debug_obsessive('* Part info before adding data:')
            debug_obsessive(pprint.pformat(part.__dict__))
            debug_obsessive('* Data found:')
            debug_obsessive(pprint.pformat(data))
            if part.datasheet is None:
                ds = next(iter(filter(lambda x: x['DocumentType'] in ['INS', 'DTE'], data['Files']['DocumentList'])), None)
                if ds:
                    part.datasheet = ds['DocumentUrl']
            specs = {}
            for sp in data.get('ParameterList', []):
                name = sp['ParameterName']
                name_l = name.lower()
                value = sp['ParameterValue']
                specs[name_l] = (name, value)
            part.update_specs(specs)
            dd = part.dd.get(DIST_NAME, DistData())
            dd.moq = data['MinAmount']
            dd.qty_increment = data['Multiples']
            dd.url = data['ProductInformationPage']
            if dd.url and dd.url.startswith('//'):
                dd.url = 'https:'+dd.url
            dd.part_num = part_stock
            dd.qty_avail = data['Amount']
            dd.currency = o.currency
            prices = data.get('PriceList', None)
            if prices:
                dd.price_tiers = {p['Amount']: p['PriceValue'] for p in prices}
            # Extra information
            dd.extra_info['desc'] = data['Description']
            value = ''
            for spec in ('capacitance', 'resistance', 'inductance'):
                val = specs.get(spec, None)
                if val:
                    value += val[1] + ' '
            if value:
                dd.extra_info['value'] = value
            for spec, name in SPEC_NAMES.items():
                val = specs.get(spec, None)
                if val:
                    dd.extra_info[name] = val[1]
            part.dd[DIST_NAME] = dd
            debug_obsessive('* Part info after adding data:')
            debug_obsessive(pprint.pformat(part.__dict__))
            # debug_obsessive(pprint.pformat(dd.__dict__))

    @staticmethod
    def query_part_info(parts, distributors, currency):
        msg = None
        try:
            api_tme._query_part_info(parts, distributors, currency)
        except TMEError as e:
            msg = e.args[0]
        if msg is not None:
            raise KiCostError(msg, ERR_SCRAPE)
        return set([DIST_NAME])


# Ok, this is special case ... we should add others
MANF_CHANGES = {'fairchild': 'onsemi'}


def _get_key(d, qty):
    """ Sorting criteria for the suggested option """
    moq = d['MinAmount']
    status = d['ProductStatusList']
    cannot_be_ordered = 'CANNOT_BE_ORDERED' in status
    hardly_available = 'HARDLY_AVAILABLE' in status
    return (cannot_be_ordered,  # Put first the ones in stock
            hardly_available,
            moq > qty)          # Put first the ones with a MOQ under the quantity we need


def _filter_by_manf(data, manf):
    """ Select the best matches according to the manufacturer """
    if not manf:
        return data
    manfs = {d['Producer'].lower() for d in data}
    if len(manfs) == 1:
        return data
    manf = manf.lower()
    best_matches = difflib.get_close_matches(manf, manfs)
    if len(best_matches) == 0:
        new_name = None
        for k, v in MANF_CHANGES.items():
            if k in manf:
                new_name = v
                break
        if new_name:
            best_matches = difflib.get_close_matches(new_name, manfs)
        if len(best_matches) == 0:
            return data
    best_match = best_matches[0]
    return list(filter(lambda x: x['Producer'].lower() == best_match, data))


def _list_comp_options(data, show, msg):
    """ Debug function used to show the list of options """
    if not show:
        return
    debug_full('  - '+msg)
    for c, d in enumerate(data):
        debug_full('  {}) {} {} moq: {} status: {}'.
                   format(c+1, d['Producer'], d['OriginalSymbol'], d['MinAmount'], d['ProductStatusList']))


def _select_best(data, manf, qty):
    """ Selects the best result """
    c = len(data)
    if c == 0:
        return None
    if c == 1:
        return data[0]['Symbol']
    debug_obsessive(' - Choosing the best match ({} options, qty: {} manf: {})'.format(c, qty, manf))
    ultra_debug = is_debug_full()
    _list_comp_options(data, ultra_debug, 'Original list')
    # Try to choose the best manufacturer
    data2 = _filter_by_manf(data, manf)
    if data != data2:
        debug_obsessive(' - Selected manf `{}`'.format(data2[0]['Producer']))
        _list_comp_options(data2, ultra_debug, 'Manufacturer selected')
        if len(data2) == 1:
            return data2[0]['Symbol']
    # Sort the results according to the best availability/price
    data3 = sorted(data2, key=lambda x: _get_key(x, qty))
    _list_comp_options(data3, ultra_debug, 'Sorted')
    return data3[0]['Symbol']


distributor_class.register(api_tme, 100)



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Hildo Guillardi Júnior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'
# This module is intended to work with Altium XML files.

# Libraries.
import sys
import os
import copy  # Necessary because Py2 doesn't have copy in list.
from datetime import datetime
from bs4 import BeautifulSoup  # To Read XML files.
import re  # Regular expression parser.
from .tools import field_name_translations, PART_REF_REGEX_NOT_ALLOWED
from .eda import eda_class
from .log__ import debug_overview
from .. import KiCostError, ERR_INPUTFILE

ALTIUM_NONE = '[NoParam]'  # Value of Altium to `None`.
ALTIUM_PART_SEPRTR = r'(?<!\\),\s*'  # Separator for the part numbers in a list, remove the lateral spaces.
FILE_REGEX = r'\<GRID[\s\S]+<COLUMNS>[\s\S]+<COLUMN[\s\S]+<\/COLUMNS>[\s\S]+<ROWS>[\s\S]+\<ROW[\s\S]+\<\/ROWS>[\s\S]+\<\/GRID>'

__all__ = ['eda_altium']


def extract_field(xml_entry, field_name):
    '''Extract XML fields from XML entry given.'''
    try:
        if sys.version_info >= (3, 0):
            return xml_entry[field_name]
        else:
            return xml_entry[field_name].encode('ascii', 'ignore')
    except KeyError:
        return None


def extract_fields_row(row, header):
    '''Extract XML fields from the part in a library or schematic.'''

    # First get the references and the quantities of elements in each row group.
    header_translated = [field_name_translations.get(hdr.lower(), hdr.lower()) for hdr in header]
    hdr_refs = [i for i, x in enumerate(header_translated) if x == "refs"]
    if not hdr_refs:
        raise KiCostError('No part designators/references found in the BOM.\nTry to generate the file again with Altium.', ERR_INPUTFILE)
    else:
        hdr_refs = hdr_refs[0]
    refs = re.split(ALTIUM_PART_SEPRTR, extract_field(row, header[hdr_refs].lower()))
    header_valid = copy.copy(header)
    header_valid.remove(header[hdr_refs])
    try:
        hdr_qty = [i for i, x in enumerate(header_translated) if x == "qty"][0]
        qty = int(extract_field(row, header[hdr_qty].lower()))
        header_valid.remove(header[hdr_qty])
        if qty != len(refs):
            raise KiCostError('Not recognize the division elements in the Altium BOM.\nIf you are using subparts, try to replace the separator from `, `'
                              ' to `,` or better, use `;` instead `,`.', ERR_INPUTFILE)
    except Exception:
        qty = len(refs)

    # After the others fields.
    fields = [dict() for x in range(qty)]
    for hdr in header_valid:
        # Extract each information, by the the header given, for each row part, spliting it in a list.
        value = extract_field(row, hdr.lower())
        value = re.split(ALTIUM_PART_SEPRTR, value)
        for i in range(qty):
            if len(value) == qty:
                v = value[i]
            else:
                v = value[0]  # Footprint is just one for group.
            fields[i][field_name_translations.get(hdr.lower(), hdr.lower())] = v
    return refs, fields


def get_part_groups(in_file):
    '''@brief Get groups of identical parts from an XML file and return them as a dictionary.
       @param in_file `str()` with the file name.
       @return `dict()` of the parts designed. The keys are the componentes references.
    '''
    # Read-in the schematic XML file to get a tree and get its root.
    debug_overview('# Getting from XML \'{}\' Altium BoM...'.format(
                                    os.path.basename(in_file)))
    file_h = open(in_file)
    root = BeautifulSoup(file_h, 'xml')
    file_h.close()

    # Get the header of the XML file of Altium, so KiCost is able to to
    # to get all the informations in the file.
    debug_overview('Getting the XML table header...')
    header = [extract_field(entry, 'name') for entry in root.find('columns').find_all('column')]

    debug_overview('Getting components...')
    accepted_components = {}
    for row in root.find('rows').find_all('row'):

        # Get the values for the fields in each library part (if any).
        refs, fields = extract_fields_row(row, header)
        for i in range(len(refs)):
            ref = refs[i]
            ref = re.sub(r'\+$', 'p', ref)  # Finishing "+".
            ref = re.sub(PART_REF_REGEX_NOT_ALLOWED, '', ref)  # Generic special characters not allowed. To work around #ISSUE #89.
            ref = re.sub(r'\-+', '-', ref)  # Double "-".
            ref = re.sub(r'^\-', '', ref)  # Starting "-".
            ref = re.sub(r'\-$', 'n', ref)  # Finishing "-".
            if not re.search(r'\d$', ref):
                ref += '0'
            accepted_components[re.sub(PART_REF_REGEX_NOT_ALLOWED, '', ref)] = fields[i]

    # Not founded project information at the file content.
    prj_info = {'title': os.path.basename(in_file),
                'company': None,
                'date': datetime.fromtimestamp(os.path.getmtime(in_file)).strftime("%Y-%m-%d %H:%M:%S") + ' (file)'}

    return accepted_components, prj_info


class eda_altium(eda_class):
    name = 'altium'
    label = 'Altium file'  # Label used on the GUI.
    desc = 'Altium Limited (formerly known as Protel until 2001).'

    @staticmethod
    def get_part_groups(in_file, distributors):
        return get_part_groups(in_file)

    @staticmethod
    def file_eda_match(content, extension):
        ''' Returns True if this EDA can handle this file. '''
        return extension == '.xml' and re.search(FILE_REGEX, content, re.IGNORECASE)


eda_class.register(eda_altium)



#!/usr/bin/python3
"""
Very simple HTTP server in python (Updated for Python 3.7)

Usage:

    ./dummy-web-server.py -h
    ./dummy-web-server.py -l localhost -p 8000

Send a GET request:

    curl http://localhost:8000

Send a HEAD request:

    curl -I http://localhost:8000

Send a POST request:

    curl -d 'foo=bar&bin=baz' http://localhost:8000

This code is available for use under the MIT license.

----

Copyright 2021 Brad Montgomery

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the 'Software'), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial
portions of the Software.

THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
import argparse
import json
import os.path as op
import sys
from urllib.parse import unquote
from http.server import HTTPServer, BaseHTTPRequestHandler
queries = {}
comments = {}


class S(BaseHTTPRequestHandler):
    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def _html(self, message):
        """This just generates an HTML document that includes `message`
        in the body. Override, or re-write this do do more interesting stuff.

        """
        content = "<html><body><h1>{}</h1></body></html>".format(message)
        return content.encode("utf8")  # NOTE: must return a bytes object!

    def do_GET(self):
        get_data = unquote(self.path)[2:]
        self._set_headers()
        print('Unknown query, {}'.format(get_data))
        content = "<html><body><h1>GET!</h1><pre>{}</pre></body></html>".format(get_data)
        self.wfile.write(content.encode("utf8"))
        sys.stdout.flush()

    def do_HEAD(self):
        self._set_headers()

    def do_POST(self):
        content_length = int(self.headers['Content-Length'])  # <--- Gets the size of data
        post_data = self.rfile.read(content_length).decode('utf8')  # <--- Gets the data itself
        self._set_headers()
        if post_data.startswith('{"query": '):
            # Nexar hack
            data = json.loads(post_data)
            post_data = 'query={}&variables={}'.format(data["query"], data["variables"])
        if post_data in queries:
            self.wfile.write(queries[post_data].encode("utf8"))
            print("Known query "+comments[post_data])
        else:
            post_data = post_data.replace('%7E', '~')
            if post_data in queries:
                self.wfile.write(queries[post_data].encode("utf8"))
                print("Known query (7E replaced) "+comments[post_data])
            else:
                data = unquote(post_data.replace('+', ' '))
                print('Unknown query, len={}\n{}\n{}'.format(content_length, post_data, data))
                content = "<html><body><h1>POST!</h1><pre>{}</pre></body></html>".format(post_data)
                self.wfile.write(content.encode("utf8"))
        sys.stdout.flush()


def load_queries(file):
    global queries
    with open(file, 'rt') as f:
        is_query = True
        last_comment = None
        id = 1
        for line in f:
            line = line[:-1]
            if line[0] == '#':
                last_comment = line[1:].strip()
                id = 1
            elif is_query:
                query = line
                # Octopart changed it. It was before we collected the queries.
                # To avoid collecting all of them we currently patch the queries.
                query = query.replace('Arrow+Electronics%2C+Inc.', 'Arrow+Electronics')
                is_query = False
            else:
                # print(f"Read: {query}")
                # print(unquote(query.replace('+', ' ')))
                # print(len(query))
                if query.startswith('query=query+MultiMatchSearch'):
                    # Nexar hack
                    # Not good, nor elegant, but recycles what we have
                    query = unquote(query.replace('+', ' '))
                    # print(f"Translated: {query}")
                queries[query] = line
                comments[query] = '{} ({})'.format(last_comment, id)
                id += 1
                is_query = True


def run(server_class=HTTPServer, handler_class=S, addr="localhost", port=8000):
    server_address = (addr, port)
    httpd = server_class(server_address, handler_class)
    load_queries(op.join(op.dirname(__file__), 'kitspace_queries.txt'))

    print("Starting httpd server on {}:{}".format(addr, port))
    httpd.serve_forever()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run a simple HTTP server")
    parser.add_argument(
        "-l",
        "--listen",
        default="localhost",
        help="Specify the IP address on which the server listens",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8000,
        help="Specify the port on which the server listens",
    )
    args = parser.parse_args()
    run(addr=args.listen, port=args.port)



# -*- coding: utf-8 -*-

__author__ = 'XESS Corporation'
__email__ = 'info@xess.com'
# Export .version.__version__ as a module version
from .version import __version__, __build__  # noqa: F401

# Definition used inside KiCost and its submodules, here for convenience
from .log__ import (debug, debug_detailed, is_debug_detailed, debug_overview,  is_debug_overview, debug_obsessive,  is_debug_obsessive,  # noqa F401
                    debug_full, debug_general, info, warning, error, get_logger, set_logger, DEBUG_OVERVIEW, DEBUG_DETAILED,
                    DEBUG_OBSESSIVE, DEBUG_HTTP_HEADERS, DEBUG_HTTP_RESPONSES, DEBUG_FULL)
from .global_vars import (SEPRTR, W_FLDOVR, ERR_INPUTFILE, W_DUPWRONG, ERR_FIELDS, W_INCQTY, W_REPMAN, W_MANQTY, W_NOINFO, ERR_SCRAPE,   # noqa: F401
                          W_APIFAIL, W_AMBIPN, W_ASSQTY, DEFAULT_CURRENCY, NO_PRICE, W_BADPRICE, BASE_OP_TYPES)

# Definitions exported to other applications using KiCost as a module
from .global_vars import KiCostError, DistData, PartGroup  # noqa F401
from .kicost import query_part_info, solve_parts_qtys  # noqa F401
from .config import load_config  # noqa F401
from .spreadsheet import create_worksheet, Spreadsheet  # noqa F401
from .__main__ import configure_kicost_apis, ProgressConsole, init_all_loggers  # noqa F401
from .distributors import get_distributors_list, get_dist_name_from_label, set_distributors_progress, is_valid_api  # noqa F401



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2021 by Salvador E. Tropea / Instituto Nacional de Tecnologia Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft'
__company__ = 'Instituto Nacional de Tecnologia Industrial - Argentina'

# Libraries.
import pprint
import requests
import difflib

# KiCost definitions.
from .. import KiCostError, DistData, W_NOINFO, ERR_SCRAPE, W_APIFAIL, W_AMBIPN
# Distributors definitions.
from .distributor import distributor_class, QueryCache, hide_secrets
from .log__ import debug_detailed, debug_overview, debug_obsessive, debug_full, warning, is_debug_full

# Countries supported by Farnell (*.farnell.com)
FARNELL_SUP = {'bg': 'EUR',
               'cz': 'CZK',
               'dk': 'DKK',
               'at': 'EUR',
               'ch': 'CHF',
               'de': 'EUR',
               'ie': 'EUR',
               'il': 'USD',
               'uk': 'GBP',
               'es': 'EUR',
               'ee': 'EUR',
               'fi': 'EUR',
               'fr': 'EUR',
               'hu': 'HUF',
               'it': 'EUR',
               'lt': 'EUR',
               'lv': 'EUR',
               'be': 'EUR',
               'nl': 'EUR',
               'no': 'NOK',
               'pl': 'PLN',
               'pt': 'EUR',
               'ro': 'RON',
               'ru': 'EUR',
               'sk': 'EUR',
               'si': 'EUR',
               'se': 'SEK',
               'tr': 'EUR'}
# Countries supported by Newark
NEWARK_SUP = {'us': ('www.newark.com', 'USD'),
              'ca': ('canada.newark.com', 'CAD'),
              'mx': ('mexico.newark.com', 'USD')}
# Countries supported by Element14, they accept Newark and Farnell SKUs
ELEMENT14_SUP = {'cn': 'CNY',
                 'au': 'AUD',
                 'nz': 'NZD',
                 'hk': 'HKD',
                 'sg': 'SGD',
                 'my': 'MYR',
                 'ph': 'PHP',
                 'th': 'THB',
                 'in': 'INR',
                 'kr': 'KRW',
                 'vn': 'USD',
                 # The European Central Bank doesn't provide exchange for Taiwan
                 # 'tw': 'TWD'
                 }
# Countries supported by CPC
CPC_SUP = {'uk': ('cpc.farnell.com', 'GBP'),
           'ie': ('cpcireland.farnell.com', 'EUR')}
# Specs known by KiCost
SPEC_NAMES = {'power rating': 'power',
              'voltage rating': 'voltage',
              'vendorName': 'manf',
              'temperature coefficient': 'temp_coeff'}

# DIST_NAMES = ['cpc', 'farnell', 'newark']
DIST_NAMES = ['farnell', 'newark']
PRE_TERM = {'sku': 'id', 'key': 'any', 'mpn': 'manuPartNum'}
REPLY = {'sku': 'premierFarnellPartNumberReturn', 'key': 'keywordSearchReturn', 'mpn': 'manufacturerPartNumberSearchReturn'}

BASE_URL = 'https://api.element14.com/catalog/products'

__all__ = ['api_element14']


class Element14Error(Exception):
    pass


class Element14(object):
    def __init__(self, dist, country, key, cache):
        if dist == 'farnell':
            # Farnell Europe
            cur = FARNELL_SUP.get(country, None)
            if cur is not None:
                self.currency = cur
                self.store_info_id = country + '.farnell.com'
            else:
                # Element14 Asia and Oceania
                cur = ELEMENT14_SUP.get(country, None)
                if cur is not None:
                    self.currency = cur
                    self.store_info_id = country + '.element14.com'
                else:
                    raise Element14Error('Unsupported country `{}` for `{}`'.format(country, dist))
        elif dist == 'newark':
            # Newark North America
            url, cur = NEWARK_SUP.get(country, (None, None))
            if cur is not None:
                self.currency = cur
                self.store_info_id = url
            else:
                # Element14 Asia and Oceania
                cur = ELEMENT14_SUP.get(country, None)
                if cur is not None:
                    self.currency = cur
                    self.store_info_id = country + '.element14.com'
                else:
                    raise Element14Error('Unsupported country `{}` for `{}`'.format(country, dist))
        elif dist == 'cpc':
            # Great Britain CPC
            url, cur = CPC_SUP.get(country, (None, None))
            if cur is not None:
                self.currency = cur
                self.store_info_id = url
            else:
                raise Element14Error('Unsupported country `{}` for `{}`'.format(country, dist))
        self.key = key
        self.cache = cache

    def _extract_data(self, data, kind, term):
        res = REPLY[kind]
        if res not in data:
            raise Element14Error("Malformed reply " + str(data))
        data = data[res]
        c = None
        try:
            c = data['numberOfResults']
        except KeyError:
            pass
        if c is None:
            raise Element14Error("Missing `numberOfResults` " + str(data))
        if c == 0:
            return None
        if 'products' not in data:
            raise Element14Error("Missing `products` " + str(data))
        prod = data['products']
        if c != 1:
            warning(W_AMBIPN, "Got {} hits for {} {}".format(c, kind, term))
        return prod

    def search(self, kind, term, part={}):
        # Try to get the data from the cache
        full_name = term+'_'+self.store_info_id
        data, loaded = self.cache.load_results(kind, full_name)
        if loaded:
            debug_obsessive('Data from cache: '+pprint.pformat(data))
            return self._extract_data(data, kind, term)
        # Do a query
        params = {'callInfo.responseDataFormat': 'JSON',
                  'term': PRE_TERM[kind]+':'+term,
                  'storeInfo.id': self.store_info_id,
                  'callInfo.apiKey': self.key,
                  'resultsSettings.responseGroup': 'large'}
        if kind == 'key':
            params['resultsSettings.offset'] = '0'
            params['resultsSettings.numberOfResults'] = '10'
        params.update(part)
        key_hide = hide_secrets(self.key)
        debug_obsessive('Query params: '+pprint.pformat(params).replace(self.key, key_hide))
        r = requests.get(BASE_URL, params=params)
        if r.status_code != 200:
            # debug_obsessive(pprint.pformat(r.__dict__))
            raise Element14Error("Server error `{}` ({})".format(r.reason, r.status_code))
        data = r.json()
        debug_obsessive('Data from server: '+pprint.pformat(data))
        self.cache.save_results(kind, full_name, data)
        return self._extract_data(data, kind, term)

    def by_sku(self, sku):
        debug_detailed('Search by SKU '+sku)
        return self.search('sku', sku)

    def by_keyword(self, key):
        debug_detailed('Search by keyword '+key)
        return self.search('key', key)

    def by_manf_pn(self, pn):
        debug_detailed('Search by part number '+pn)
        return self.search('mpn', pn)


class api_element14(distributor_class):
    name = 'Element14'
    type = 'api'
    # Currently enabled only by request
    enabled = True
    url = 'https://partner.element14.com/'  # Web site API information.
    api_distributors = DIST_NAMES
    # Options supported by this API
    config_options = {'key': str,
                      'try_by_keyword': bool,
                      'farnell_country': ('BG', 'CZ', 'DK', 'AT', 'CH', 'DE', 'IE', 'IL', 'UK', 'ES', 'EE', 'FI', 'FR', 'HU', 'IT', 'LT',
                                          'LV', 'BE', 'NL', 'NO', 'PL', 'PT', 'RO', 'RU', 'SK', 'SI', 'SE', 'TR', 'CN', 'AU', 'NZ', 'HK',
                                          'SG', 'MY', 'PH', 'TH', 'IN', 'KR', 'VN'),
                      'newark_country': ('CA', 'US', 'MX', 'CN', 'AU', 'NZ', 'HK', 'SG', 'MY', 'PH', 'TH', 'IN', 'KR', 'VN'),
                      'cpc_country': ('IE', 'UK')}
    key = None
    countries = {'cpc': 'uk', 'farnell': 'uk', 'newark': 'us'}
    try_by_keyword = False
    env_prefix = 'ELEMENT14'
    env_ops = {'ELEMENT14_PART_API_KEY': 'key'}

    @staticmethod
    def configure(ops):
        cache_ttl = 7
        cache_path = None
        for k, v in ops.items():
            if k == 'key':
                api_element14.key = v
            elif k == 'enable':
                api_element14.enabled = v
            elif k == 'try_by_keyword':
                api_element14.try_by_keyword = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
            elif k.endswith('_country'):
                api_element14.countries[k[:-8]] = v.lower()
        if api_element14.enabled and api_element14.key is None:
            warning(W_APIFAIL, "Can't enable Elemen14 without a `key`")
            api_element14.enabled = False
        debug_obsessive('Element14 API configured to enabled {} key {} path {}'.
                        format(api_element14.enabled, hide_secrets(api_element14.key), cache_path))
        if not api_element14.enabled:
            return
        # Configure the cache
        api_element14.cache = QueryCache(cache_path, cache_ttl)

    @staticmethod
    def _query_part_info(dist, country, parts, distributors, currency):
        '''Fill-in the parts with price/qty/etc info from KitSpace.'''
        debug_overview('# Getting part data from Element14 ({} {})...'.format(dist, country))
        field_cat = dist + '#'
        o = Element14(dist, country, api_element14.key, api_element14.cache)

        # Setup progress bar to track progress of server queries.
        progress = distributor_class.progress(len(parts), distributor_class.logger)
        for part in parts:
            data = None
            # Get the Element14 P/N for this part
            part_stock = part.fields.get(field_cat)
            part_manf = part.fields.get('manf', '')
            part_code = part.fields.get('manf#')
            if part_stock:
                debug_detailed('\n**** {} P/N: {}'.format(dist, part_stock))
                data = o.by_sku(part_stock)
                if data is None:
                    warning(W_NOINFO, 'The \'{}\' {} code is not valid'.format(part_stock, dist))
                    if api_element14.try_by_keyword:
                        data = o.by_keyword(part_stock)
            else:
                # No Element14 P/N, search using the manufacturer code
                if part_code:
                    if part_manf:
                        debug_detailed('\n**** Manufacturer: {} P/N: {}'.format(part_manf, part_code))
                    else:
                        debug_detailed('\n**** P/N: {}'.format(part_code))
                    data = o.by_manf_pn(part_code)
                    if data is None and api_element14.try_by_keyword:
                        data = o.by_keyword(part_code)
            if data is None:
                warning(W_NOINFO, 'No information found at {} for part/s \'{}\''.format(dist, part.refs))
            else:
                data = _select_best(data, part_manf, part.qty_total_spreadsheet)
                debug_obsessive('* Part info before adding data:')
                debug_obsessive(pprint.pformat(part.__dict__))
                debug_obsessive('* Data found:')
                debug_obsessive(pprint.pformat(data))
                ds = data.get('datasheets', None)
                if part.datasheet is None and ds is not None:
                    part.datasheet = ds[0]['url']
                if part.lifecycle is None:
                    part.lifecycle = 'obsolete' if data['productStatus'] == 'NO_LONGER_MANUFACTURED' else 'active'
                tolerance = footprint = frequency = None
                specs = {}
                if 'rohsStatusCode' in data:
                    specs['rohs'] = ('RoHS', data['rohsStatusCode'])
                for sp in data.get('attributes', []):
                    name = sp['attributeLabel']
                    name_l = name.lower()
                    value = sp['attributeValue']
                    unit = sp.get('attributeUnit', None)
                    if unit:
                        value = value + ' ' + unit
                    specs[name_l] = (name, value)
                    if name_l.endswith('tolerance'):
                        tolerance = value
                    if name_l.endswith('case style'):
                        footprint = value
                    if name_l.endswith('frequency'):
                        frequency = value
                part.update_specs(specs)
                dd = part.dd.get(dist, DistData())
                dd.qty_increment = dd.moq = data['translatedMinimumOrderQuality']
                dd.url = 'https://'+o.store_info_id+'/w/search?st='+data['sku']
                dd.part_num = data['sku']
                dd.qty_avail = data['inv']
                dd.currency = o.currency
                prices = data.get('prices', None)
                if prices:
                    dd.price_tiers = {p['from']: p['cost'] for p in prices}
                # Extra information
                dd.extra_info['desc'] = data['displayName']
                value = ''
                for spec in ('capacitance', 'resistance', 'inductance'):
                    val = specs.get(spec, None)
                    if val:
                        value += val[1] + ' '
                if value:
                    dd.extra_info['value'] = value
                if tolerance:
                    dd.extra_info['tolerance'] = tolerance
                if footprint:
                    dd.extra_info['footprint'] = footprint
                if frequency:
                    dd.extra_info['frequency'] = frequency
                for spec, name in SPEC_NAMES.items():
                    val = specs.get(spec, None)
                    if val:
                        dd.extra_info[name] = val[1]
                part.dd[dist] = dd
                debug_obsessive('* Part info after adding data:')
                debug_obsessive(pprint.pformat(part.__dict__))
                # debug_obsessive(pprint.pformat(dd.__dict__))
            progress.update(1)
        progress.close()

    @staticmethod
    def query_part_info(parts, distributors, currency):
        if len(set(DIST_NAMES).intersection(distributors)) == 0:
            # None of our distributors is used
            debug_overview('# Skipping Element14 plug-in')
            return set()
        msg = None
        try:
            for dist in DIST_NAMES:
                country = api_element14.countries[dist]
                api_element14._query_part_info(dist, country, parts, distributors, currency)
        except Element14Error as e:
            msg = e.args[0]
        if msg is not None:
            raise KiCostError(msg, ERR_SCRAPE)
        return set(DIST_NAMES)


# Ok, this is special case ... we should add others
MANF_CHANGES = {'fairchild': 'onsemi'}


def _get_price(d):
    """ Return the first price or -1 if none """
    prices = d.get('prices', None)
    return prices[0]['cost'] if prices else -1


def _get_key(d, qty):
    """ Sorting criteria for the suggested option """
    price = _get_price(d)
    stock = d['inv']
    moq = d['translatedMinimumOrderQuality']
    return (price == -1,    # Put first the ones with price
            stock == 0,     # Put first the ones in stock
            stock < qty,    # Put first the ones with enough stock
            # TODO: some tollerance to the MOQ? Like buy 20% extra
            moq > qty,      # Put first the ones with a MOQ under the quantity we need
            d['productStatus'].startswith('NO_LONGER_'),   # Put first the active components
            price,
            moq)  # At equal price suggest the lower MOQ


def _filter_by_manf(data, manf):
    """ Select the best matches according to the manufacturer """
    if not manf:
        return data
    manfs = {d['brandName'].lower() for d in data}
    if len(manfs) == 1:
        return data
    manf = manf.lower()
    best_matches = difflib.get_close_matches(manf, manfs)
    if len(best_matches) == 0:
        new_name = None
        for k, v in MANF_CHANGES.items():
            if k in manf:
                new_name = v
                break
        if new_name:
            best_matches = difflib.get_close_matches(new_name, manfs)
        if len(best_matches) == 0:
            return data
    best_match = best_matches[0]
    return list(filter(lambda x: x['brandName'].lower() == best_match, data))


def _list_comp_options(data, show, msg):
    """ Debug function used to show the list of options """
    if not show:
        return
    debug_full('  - '+msg)
    for c, d in enumerate(data):
        debug_full('  {}) {} {} inv: {} moq: {} status: {} price: {}'.
                   format(c+1, d['brandName'], d['translatedManufacturerPartNumber'], d['inv'],
                          d['translatedMinimumOrderQuality'], d['productStatus'], _get_price(d)))


def _select_best(data, manf, qty):
    """ Selects the best result """
    c = len(data)
    if c == 1:
        return data[0]
    debug_obsessive(' - Choosing the best match ({} options, qty: {} manf: {})'.format(c, qty, manf))
    ultra_debug = is_debug_full()
    _list_comp_options(data, ultra_debug, 'Original list')
    # Try to choose the best manufacturer
    data2 = _filter_by_manf(data, manf)
    if data != data2:
        debug_obsessive(' - Selected manf `{}`'.format(data2[0]['brandName']))
        _list_comp_options(data2, ultra_debug, 'Manufacturer selected')
        if len(data2) == 1:
            return data2[0]
    # Sort the results according to the best availability/price
    data3 = sorted(data2, key=lambda x: _get_key(x, qty))
    _list_comp_options(data3, ultra_debug, 'Sorted')
    return data3[0]


distributor_class.register(api_element14, 100)



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2021 by Salvador E. Tropea / Instituto Nacional de Tecnologia Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft'
__company__ = 'Instituto Nacional de Tecnologia Industrial - Argentina'

# Libraries.
import pprint

# KiCost definitions.
from .. import DistData, KiCostError, W_NOINFO, ERR_SCRAPE, W_APIFAIL
# Distributors definitions.
from .distributor import distributor_class, QueryCache, hide_secrets
from .log__ import debug_detailed, debug_overview, debug_obsessive, warning

available = True
try:
    from kicost_digikey_api_v3 import by_digikey_pn, by_manf_pn, by_keyword, DigikeyError, DK_API
except ImportError:
    available = False

    class DK_API(object):
        api_ops = None

DIST_NAME = 'digikey'
# Specs known by KiCost
SPEC_NAMES = {'tolerance': 'tolerance',
              'power (watts)': 'power',
              'voltage - rated': 'voltage',
              'manufacturer': 'manf',
              'size / dimension': 'size',
              'temperature coefficient': 'temp_coeff',
              'frequency': 'frequency',
              'package / case': 'footprint'}
__all__ = ['api_digikey']


class api_digikey(distributor_class):
    name = 'Digi-Key'
    type = 'api'
    # Currently enabled only by request
    enabled = available
    url = 'https://developer.digikey.com/'  # Web site API information.
    api_distributors = [DIST_NAME]
    # Options supported by this API
    config_options = {'client_id': str,
                      'client_secret': str,
                      'exclude_market_place_products': bool,
                      'sandbox': bool,
                      'locale_site': ('US', 'CA', 'JP', 'UK', 'DE', 'AT', 'BE', 'DK', 'FI', 'GR', 'IE',
                                      'IT', 'LU', 'NL', 'NO', 'PT', 'ES', 'KR', 'HK', 'SG', 'CN', 'TW',
                                      'AU', 'FR', 'IN', 'NZ', 'SE', 'MX', 'CH', 'IL', 'PL', 'SK', 'SI',
                                      'LV', 'LT', 'EE', 'CZ', 'HU', 'BG', 'MY', 'ZA', 'RO', 'TH', 'PH'),
                      'locale_language': ('en', 'ja', 'de', 'fr', 'ko', 'zhs,' 'zht', 'it', 'es', 'he',
                                          'nl', 'sv', 'pl', 'fi', 'da', 'no'),
                      'locale_currency': ('USD', 'CAD', 'JPY', 'GBP', 'EUR', 'HKD', 'SGD', 'TWD', 'KRW',
                                          'AUD', 'NZD', 'INR', 'DKK', 'NOK', 'SEK', 'ILS', 'CNY', 'PLN',
                                          'CHF', 'CZK', 'HUF', 'RON', 'ZAR', 'MYR', 'THB', 'PHP'),
                      'locale_ship_to_country': str}
    # Legacy environment mapping, others will be automatically filled by `register`
    env_prefix = 'DIGIKEY'
    env_ops = {'DIGIKEY_STORAGE_PATH': 'cache_path',
               'DIGIKEY_CLIENT_SANDBOX': 'sandbox'}

    @staticmethod
    def configure(ops):
        DK_API.api_ops = {}
        cache_ttl = 7
        cache_path = None
        if not available:
            debug_obsessive('Digi-Key API not available')
            return
        for k, v in ops.items():
            if k == 'client_id':
                DK_API.id = v
            elif k == 'client_secret':
                DK_API.secret = v
            elif k == 'enable' and available:
                api_digikey.enabled = v
            elif k == 'sandbox':
                DK_API.sandbox = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
            elif k == 'exclude_market_place_products':
                DK_API.exclude_market_place_products = v
            elif k.startswith('locale_'):
                DK_API.api_ops[k] = v
        if api_digikey.enabled and (DK_API.id is None or DK_API.secret is None or cache_path is None):
            warning(W_APIFAIL, "Can't enable Digi-Key without a `client_id`, `client_secret` and `cache_path`")
            api_digikey.enabled = False
        debug_obsessive('Digi-Key API configured to enabled {} id {} secret {} path {}'.
                        format(api_digikey.enabled, hide_secrets(DK_API.id), hide_secrets(DK_API.secret), cache_path))
        if not api_digikey.enabled:
            return
        # Try to configure the plug-in
        cache = QueryCache(cache_path, cache_ttl)
        try:
            DK_API.configure(cache, a_logger=distributor_class.logger)
        except DigikeyError as e:
            warning(W_APIFAIL, 'Failed to init Digi-Key API, reason: {}'.format(e.args[0]))
            api_digikey.enabled = False

    @staticmethod
    def _query_part_info(parts, distributors, currency):
        '''Fill-in the parts with price/qty/etc info from KitSpace.'''
        if DIST_NAME not in distributors:
            debug_overview('# Skipping Digi-Key plug-in')
            return
        debug_overview('# Getting part data from Digi-Key...')
        field_cat = DIST_NAME + '#'

        # Setup progress bar to track progress of server queries.
        progress = distributor_class.progress(len(parts), distributor_class.logger)
        for part in parts:
            data = None
            # Get the Digi-Key P/N for this part
            part_stock = part.fields.get(field_cat)
            if part_stock:
                debug_detailed('\n**** Digi-Key P/N: {}'.format(part_stock))
                o = by_digikey_pn(part_stock)
                data = o.search()
                if data is None:
                    warning(W_NOINFO, 'The \'{}\' Digi-Key code is not valid'.format(part_stock))
                    o = by_keyword(part_stock)
                    data = o.search()
            else:
                # No Digi-Key P/N, search using the manufacturer code
                part_manf = part.fields.get('manf', '')
                part_code = part.fields.get('manf#')
                if part_code:
                    if part_manf:
                        debug_detailed('\n**** Manufacturer: {} P/N: {}'.format(part_manf, part_code))
                    else:
                        debug_detailed('\n**** P/N: {}'.format(part_code))
                    o = by_manf_pn(part_code)
                    data = o.search()
                    if data is None:
                        o = by_keyword(part_code)
                        data = o.search()
            if data is None:
                warning(W_NOINFO, 'No information found at Digi-Key for part/s \'{}\''.format(part.refs))
            else:
                debug_obsessive('* Part info before adding data:')
                debug_obsessive(pprint.pformat(part.__dict__))
                debug_obsessive('* Data found:')
                debug_obsessive(str(data))
                part.datasheet = data.primary_datasheet
                part.lifecycle = data.product_status.lower()
                specs = {sp.parameter.lower(): (sp.parameter, sp.value) for sp in data.parameters}
                specs['rohs'] = ('RoHS', data.ro_hs_status)
                part.update_specs(specs)
                dd = part.dd.get(DIST_NAME, DistData())
                dd.qty_increment = dd.moq = data.minimum_order_quantity
                dd.url = data.product_url
                dd.part_num = data.digi_key_part_number
                dd.qty_avail = data.quantity_available
                dd.currency = data.search_locale_used.currency
                dd.price_tiers = {p.break_quantity: p.unit_price for p in data.standard_pricing}
                # Extra information
                if data.product_description:
                    dd.extra_info['desc'] = data.product_description
                value = ''
                for spec in ('capacitance', 'resistance', 'inductance'):
                    val = specs.get(spec, None)
                    if val:
                        value += val[1] + ' '
                if value:
                    dd.extra_info['value'] = value
                for spec, name in SPEC_NAMES.items():
                    val = specs.get(spec, None)
                    if val:
                        dd.extra_info[name] = val[1]
                part.dd[DIST_NAME] = dd
                debug_obsessive('* Part info after adding data:')
                debug_obsessive(pprint.pformat(part.__dict__))
                debug_obsessive(pprint.pformat(dd.__dict__))
            progress.update(1)
        progress.close()

    @staticmethod
    def query_part_info(parts, distributors, currency):
        msg = None
        try:
            api_digikey._query_part_info(parts, distributors, currency)
        except DigikeyError as e:
            msg = e.args[0]
        if msg is not None:
            raise KiCostError(msg, ERR_SCRAPE)
        return set([DIST_NAME])


distributor_class.register(api_digikey, 100)



# -*- coding: utf-8 -*-
# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Max Maisel / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# ******************************************************************************
# IMPORTANT: This is no longer an usable API. Is here just for testing purposes.
# ******************************************************************************

# Author information.
__author__ = 'Hildo Guillardi Júnior'
__webpage__ = 'https://github.com/hildogjr/'
__company__ = 'University of Campinas - Brazil'

# Libraries.
import json
import requests
import re
import sys
import os
import copy
import pprint
from collections import OrderedDict
if sys.version_info[0] < 3:
    from urllib import quote_plus
else:
    from urllib.parse import quote_plus

# KiCost definitions.
from .. import KiCostError, DistData, DEFAULT_CURRENCY, ERR_SCRAPE, W_NOINFO, NO_PRICE
# Distributors definitions.
from .distributor import distributor_class, QueryCache
from .log__ import debug_overview, debug_obsessive, warning


# Uncomment for debug
# Use `debug('x + 1')` for instance.
# def debug(expression):
#     frame = sys._getframe(1)
#     distributor_class.logger.info(expression, '=', repr(eval(expression, frame.f_globals, frame.f_locals)))

MAX_PARTS_PER_QUERY = 20  # Maximum number of parts in a single query.

# Information to return from PartInfo KitSpace server.

QUERY_AVAIABLE_CURRENCIES = ['GBP', 'EUR', 'USD']
# DEFAULT_CURRENCY
QUERY_ANSWER = '''
    mpn{manufacturer, part},
    datasheet,
    description,
    specs{key, value},
    offers(from: {DISTRIBUTORS}){
        product_url,
        sku {vendor, part},
        description,
        moq,
        in_stock_quantity,
        prices{''' + ','.join(QUERY_AVAIABLE_CURRENCIES) + '''}
        }
'''
# Informations not used: type,specs{key, name, value},image {url, credit_string, credit_url},stock_location
QUERY_ANSWER = re.sub(r'[\s\n]', '', QUERY_ANSWER)

QUERY_PART = 'query ($input: MpnInput!) { part(mpn: $input) {' + QUERY_ANSWER + '} }'
QUERY_MATCH = 'query ($input: [MpnOrSku]!){ match(parts: $input) {' + QUERY_ANSWER + '} }'
QUERY_SEARCH = 'query ($input: String!){ search(term: $input) {' + QUERY_ANSWER + '} }'
QUERY_URL = 'https://dev-partinfo.kitspace.org/graphql'

__all__ = ['api_partinfo_kitspace']


class QueryStruct(object):
    def __init__(self, query, part, for_dists):
        self.query = query  # Each part reference query.
        self.part = part    # Pointer to the part.
        self.for_dists = for_dists  # Used the stock code mention for disambiguation, it is used `None` for the "manf#".
        self.result = None
        self.loaded = False


class api_partinfo_kitspace(distributor_class):
    name = 'KitSpace'
    type = 'api'
    enabled = False
    url = 'https://kitspace.org/'  # Web site API information.

    api_distributors = ['digikey', 'farnell', 'mouser', 'newark', 'rs', 'arrow', 'tme', 'lcsc']
    DIST_TRANSLATION = {  # Distributor translation.
                        'Digikey': 'digikey',
                        'Farnell': 'farnell',
                        'Mouser': 'mouser',
                        'Newark': 'newark',
                        'RS': 'rs',
                        'TME': 'tme',
                        'Arrow Electronics': 'arrow',
                        'LCSC': 'lcsc',
                       }
    # Dict to translate KiCost field names into KitSpace distributor names
    KICOST2KITSPACE_DIST = {v: k for k, v in DIST_TRANSLATION.items()}
    cache = None
    env_prefix = 'KITSPACE'
    env_ops = {}

    @staticmethod
    def configure(ops):
        cache_ttl = 7
        cache_path = None
        for k, v in ops.items():
            if k == 'enable':
                api_partinfo_kitspace.enabled = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
        api_partinfo_kitspace.cache = QueryCache(cache_path, cache_ttl)
        debug_obsessive('KitSpace API configured to enabled {}'.format(api_partinfo_kitspace.enabled))

    @staticmethod
    def query(query_parts, distributors, query_type=QUERY_MATCH):
        '''Send query to server and return results.'''

        distributors = [api_partinfo_kitspace.KICOST2KITSPACE_DIST[d] for d in distributors]
        # Allow changing the URL for debug purposes
        try:
            url = os.environ['KICOST_KITSPACE_URL']
        except KeyError:
            url = QUERY_URL
        # Sort the distributors to create a reproducible query
        query_type = re.sub(r'\{DISTRIBUTORS\}', '["' + '","'.join(sorted(distributors)) + '"]', query_type)
        # r = requests.post(url, {"query": QUERY_SEARCH, "variables": variables}) #TODO future use for ISSUE #17
        variables = '{"input":[' + ','.join(query_parts) + ']}'
        # Remove all spaces, even inside the manf#
        # SET comment: this is how the code always worked. Octopart (used by KitSpace) ignores spaces inside manf# codes.
        variables = variables.replace(' ', '')
        # Do the query using POST
        data = 'query={}&variables={}'.format(quote_plus(query_type), quote_plus(variables))
        distributor_class.log_request(url, data)
        data = OrderedDict()
        data["query"] = query_type
        data["variables"] = variables
        response = requests.post(url, data)
        response.encoding = 'UTF-8'
        distributor_class.log_response(response)
        if response.status_code == requests.codes['ok']:  # 200
            results = json.loads(response.text)
            return results
        elif response.status_code == requests.codes['not_found']:  # 404
            raise KiCostError('Kitspace server not found check your internet connection.', ERR_SCRAPE)
        elif response.status_code == requests.codes['request_timeout']:  # 408
            raise KiCostError('KitSpace is not responding.', ERR_SCRAPE)
        elif response.status_code == requests.codes['bad_request']:  # 400
            raise KiCostError('Bad request to Kitspace server probably due to an incorrect string '
                              'format check your `manf#` codes and contact the suport team.', ERR_SCRAPE)
        elif response.status_code == requests.codes['gateway_timeout']:  # 504
            raise KiCostError('One of the internal Kitspace services may experiencing problems.'
                              ' Most probably is saturated, please consider using other API.', ERR_SCRAPE)
        else:
            raise KiCostError('Kitspace error: ' + str(response.status_code), ERR_SCRAPE)

    @staticmethod
    def get_spec(data, item, default=None):
        '''Get the value of `value` field of a dictionary if the `name` field identifier.
        Used to get information from the JSON response.'''
        for d in data['specs']:
            if d['key'] == item:
                value = d['value']
                return value if value is not None else default
        return default

    @staticmethod
    def query2name(query):
        ''' Finds the prefix and name for a query '''
        q = json.loads(query)
        mpn = q.get('mpn', None)
        if mpn is not None:
            prefix = 'mpn'
            name = mpn['part']
        else:
            sku = q.get('sku', None)
            if sku is not None:
                prefix = 'sku'
                name = sku['vendor'] + '_' + sku['part']
        return prefix, name

    @staticmethod
    def get_part_info(queries, distributors):
        '''Query PartInfo for quantity/price info.
           `distributors` is the list of all distributors we want, in general. '''
        only_query = [q.query for q in queries]
        results = api_partinfo_kitspace.query(only_query, distributors)
        for i, r in enumerate(results['data']['match']):
            queries[i].result = r
            # Solve the prefix and name
            prefix, name = api_partinfo_kitspace.query2name(queries[i].query)
            api_partinfo_kitspace.cache.save_results(prefix, name, r)

    @staticmethod
    def fill_part_info(queries, distributors, currency, solved):
        ''' Place the results into the parts list. '''
        # Translate from PartInfo distributor names to the names used internally by kicost.
        dist_xlate = api_partinfo_kitspace.DIST_TRANSLATION

        # Loop through the response to the query and enter info into the parts list.
        for q in queries:
            # Unpack the structure
            part_query = q.query
            part = q.part
            dist_want = q.for_dists
            result = q.result
            # Process it
            if not result:
                warning(W_NOINFO, 'No information found for parts \'{}\' query `{}`'.format(part.refs, str(part_query)))
                continue
            # Get the information of the part.
            part.datasheet = result.get('datasheet')
            part.lifecycle = api_partinfo_kitspace.get_spec(result, 'lifecycle_status', 'active').lower()
            # Misc data collected, currently not used inside KiCost
            part.update_specs({sp['key']: (sp['key'], sp['value']) for sp in result['specs'] if sp['value']})
            # Loop through the offers from various dists for this particular part.
            for offer in result['offers']:
                # Get the distributor who made the offer and add their
                # price/qty info to the parts list if its one of the accepted distributors.
                dist = dist_xlate.get(offer['sku']['vendor'], '')
                if dist not in dist_want:
                    # Not interested in this distributor
                    continue
                # Get the DistData for this distributor
                dd = part.dd.get(dist, DistData())
                # This will happen if there are not enough entries in the price/qty list.
                # As a stop-gap measure, just assign infinity to the part increment.
                # A better alternative may be to examine the packaging field of the offer.
                part_qty_increment = float("inf")
                # Get pricing information from this distributor.
                dist_currency = {cur: pri for cur, pri in offer['prices'].items() if pri}
                if not dist_currency:
                    # Some times the API returns minimum purchase 0 and a not valid `price_tiers`.
                    warning(NO_PRICE, 'No price information found for parts \'{}\' query `{}`'.format(part.refs, str(part_query)))
                else:
                    prices = None
                    # Get the price tiers prioritizing:
                    # 1) The asked currency by KiCost user;
                    # 2) The default currency given by `DEFAULT_CURRENCY` in root `global_vars.py`;
                    # 3) The first not null tiers
                    if currency in dist_currency:
                        prices = dist_currency[currency]
                        dd.currency = currency
                    elif DEFAULT_CURRENCY in dist_currency:
                        prices = dist_currency[DEFAULT_CURRENCY]
                        dd.currency = DEFAULT_CURRENCY
                    else:
                        dd.currency, prices = next(iter(dist_currency.items()))
                    price_tiers = {qty: float(price) for qty, price in prices}
                    # Combine price lists for multiple offers from the same distributor
                    # to build a complete list of cut-tape and reeled components.
                    dd.price_tiers.update(price_tiers)
                    # Compute the quantity increment between the lowest two prices.
                    # This will be used to distinguish the cut-tape from the reeled components.
                    if len(price_tiers) > 1:
                        part_break_qtys = sorted(price_tiers.keys())
                        part_qty_increment = part_break_qtys[1] - part_break_qtys[0]
                # Select the part SKU, web page, and available quantity.
                # Each distributor can have different stock codes for the same part in different
                # quantities / delivery package styles: cut-tape, reel, ...
                # Therefore we select and overwrite a previous selection if one of the
                # following conditions is met:
                #   1. We don't have a selection for this part from this distributor yet.
                #   2. The MOQ is smaller than for the current selection.
                #   3. The part_qty_increment for this offer smaller than that of the existing selection.
                #      (we prefer cut-tape style packaging over reels)
                #   4. For DigiKey, we can't use part_qty_increment to distinguish between
                #      reel and cut-tape, so we need to look at the actual DigiKey part number.
                #      This procedure is made by the definition `distributors_info[dist]['ignore_cat#_re']`
                #      at the distributor profile.
                dist_part_num = offer.get('sku', '').get('part', '')
                qty_avail = dd.qty_avail
                in_stock_quantity = offer.get('in_stock_quantity')
                if not qty_avail or (in_stock_quantity and qty_avail < in_stock_quantity):
                    # Keeps the information of more availability.
                    dd.qty_avail = in_stock_quantity  # In stock.
                ign_stock_code = distributor_class.get_distributor_info(dist).ignore_cat
                valid_part = not (ign_stock_code and re.match(ign_stock_code, dist_part_num))
                # debug('dd.part_num')  # Uncomment to debug
                # debug('dd.qty_increment')  # Uncomment to debug
                moq = offer.get('moq')
                if (valid_part and
                    (not dd.part_num or
                     (dd.qty_increment is None or part_qty_increment < dd.qty_increment) or
                     (not dd.moq or (moq and dd.moq > moq)))):
                    # Save the link, stock code, ... of the page for minimum purchase.
                    dd.moq = moq  # Minimum order qty.
                    dd.url = offer.get('product_url', '')  # Page to purchase the minimum quantity.
                    dd.part_num = dist_part_num
                    dd.qty_increment = part_qty_increment
                # Update the DistData for this distributor
                part.dd[dist] = dd
                # We have data for this distributor
                solved.add(dist)

    @staticmethod
    def query_part_info(parts, distributors, currency):
        '''Fill-in the parts with price/qty/etc info from KitSpace.'''
        debug_overview('# Getting part data from KitSpace...')

        # Use just the distributors avaliable in this API.
        # Note: The user can use --exclude and define it with fields.
        distributors = [d for d in distributors if distributor_class.get_distributor_info(d).is_web()
                        and d in api_partinfo_kitspace.api_distributors]
        FIELDS_CAT = sorted([d + '#' for d in distributors])

        # Create queries to get part price/quantities from PartInfo.
        queries = []
        # Translate from PartInfo distributor names to the names used internally by kicost.
        available_distributors = set(api_partinfo_kitspace.api_distributors)
        for part in parts:
            # Create a PartInfo query using the manufacturer's part number or the distributor's SKU.
            part_dist_use_manfpn = copy.copy(distributors)

            # Create queries using the distributor SKU
            # Check if that part have stock code that is accepted to use by this module (API).
            # KiCost will prioritize these codes under "manf#" that will be used for get
            # information for the part hat were not filled with the distributor stock code. So
            # this is checked after the 'manf#' buv code.
            found_codes_for_all_dists = True
            for d in FIELDS_CAT:
                part_stock = part.fields.get(d)
                if part_stock:
                    part_catalogue_code_dist = d[:-1]
                    if part_catalogue_code_dist in available_distributors:
                        part_code_dist = api_partinfo_kitspace.KICOST2KITSPACE_DIST[part_catalogue_code_dist]
                        query_text = '{"sku":{"vendor":"' + part_code_dist + '","part":"' + part_stock + '"}}'
                        queries.append(QueryStruct(query_text, part, [part_catalogue_code_dist]))
                        part_dist_use_manfpn.remove(part_catalogue_code_dist)
                else:
                    found_codes_for_all_dists = False

            # Create a query using the manufacturer P/N
            part_manf = part.fields.get('manf', '')
            part_code = part.fields.get('manf#')
            if part_code and not found_codes_for_all_dists:
                # Not all distributors has code, add a query for the manufaturer P/N
                query_text = '{"mpn":{"manufacturer":"' + part_manf + '","part":"' + part_code + '"}}'
                # List of distributors without an specific part number
                queries.append(QueryStruct(query_text, part, part_dist_use_manfpn))

        n_queries = len(queries)
        debug_overview('Queries {}'.format(n_queries))
        if not n_queries:
            return

        # Try to solve the queries from the cache
        unsolved = []
        for query in queries:
            # Solve the prefix and name
            prefix, name = api_partinfo_kitspace.query2name(query.query)
            # Look in the cache
            query.result, query.loaded = api_partinfo_kitspace.cache.load_results(prefix, name)
            if not query.loaded:
                unsolved.append(query)
            else:
                debug_obsessive('Data from cache: '+pprint.pformat(query.result))

        # Solve the rest from the site
        n_unsolved = len(unsolved)
        debug_overview('Cached entries {} (of {})'.format(n_queries-n_unsolved, n_queries))
        if n_unsolved:
            # Setup progress bar to track progress of server queries.
            progress = distributor_class.progress(n_unsolved, distributor_class.logger)

            # Slice the queries into batches of the largest allowed size and gather
            # the part data for each batch.
            for i in range(0, n_unsolved, MAX_PARTS_PER_QUERY):
                slc = slice(i, i+MAX_PARTS_PER_QUERY)
                api_partinfo_kitspace.get_part_info(unsolved[slc], distributors)
                progress.update(len(unsolved[slc]))

            # Done with the scraping progress bar so delete it or else we get an
            # error when the program terminates.
            progress.close()

        # Transfer the results
        solved_dist = set()
        api_partinfo_kitspace.fill_part_info(queries, distributors, currency, solved_dist)
        return solved_dist


distributor_class.register(api_partinfo_kitspace, 50)



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2018 by XESS Corporation / Hildo Guillardi Júnior
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = 'XESS Corporation'
__email__ = 'info@xess.com'

from .distributor import distributor_class, hide_secrets  # noqa: F401
# Export the ORDER_COL_USERFIELDS content
from .distributors_info import ORDER_COL_USERFIELDS  # noqa: F401

# Import and register here the API / local / scrape modules.
from .dist_local_template import dist_local_template  # noqa: F401
from .api_partinfo_kitspace import api_partinfo_kitspace  # noqa: F401
from .api_digikey import api_digikey  # noqa: F401
from .api_mouser import api_mouser  # noqa: F401
from .api_element14 import api_element14  # noqa: F401
from .api_tme import api_tme  # noqa: F401
from .api_nexar import api_nexar  # noqa: F401


#
# Some wrappers
#
def init_distributor_dict():
    distributor_class.main_init_dist_dict()


def get_dist_parts_info(parts, dist_list, currency):
    distributor_class.get_dist_parts_info(parts, dist_list, currency)


def get_registered_apis():
    return distributor_class.registered


def get_distributors_list():
    ''' List of distributors registered by the API modules '''
    return list(distributor_class.get_distributors_iter())


def get_distributors_iter():
    ''' Iterator for the distributors registered by the API modules '''
    return distributor_class.get_distributors_iter()


def get_distributor_info(name):
    ''' Gets all the information about a supported distributor.
        This information comes from the list collected from the APIs, not from the fixed template. '''
    return distributor_class.get_distributor_info(name)


def get_dist_name_from_label(label):
    ''' Returns the internal distributor name for a provided label. '''
    return distributor_class.label2name.get(label.lower())


def set_distributors_logger(logger):
    ''' Sets the logger used by the class '''
    distributor_class.logger = logger


def set_distributors_progress(cls):
    ''' Configures the class used to indicate progress '''
    distributor_class.progress = cls


def configure_apis(options):
    ''' Configure all APIs. options is a dict API -> api_options '''
    distributor_class.configure_apis(options)


def set_api_status(api, enabled):
    ''' Enable/Disable a particular API '''
    distributor_class.set_api_status(api, enabled)


def get_api_status(api):
    ''' Find if an API is enabled '''
    return distributor_class.get_api_status(api)


def is_valid_api(api):
    ''' Determines if this API is registered '''
    return distributor_class._get_api(api) is not None


def get_api_list():
    ''' Returns a list of registered APIs '''
    return [api.name for api in distributor_class.registered]


def get_api_valid_options():
    ''' Returns the vali options for each API '''
    return {api.name: api.config_options for api in distributor_class.registered}


def get_api_keys_to_hide(api):
    ''' Returns a set of options that should be hidden '''
    return distributor_class._get_api(api).keys_to_hide


def configure_from_environment(options, overwrite):
    distributor_class.configure_from_environment(options, overwrite)



# -*- coding: utf-8 -*-
# Nexar API implementation, replaces Octopart API
#
# MIT license
#
# Copyright (C) 2022 by Salvador E. Tropea / Instituto Nacional de Tecnologia Industrial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# I took the basic query from  https://github.com/NexarDeveloper/nexar-first-supply-query.git
# and the example in Nexar's IDE
# The class is based on the api_partinfo_kitspace.py class
#
# Important note:
# We get offers from tons of distributors, there is no way to say which ones we want.
# In:
# https://support.nexar.com/support/solutions/articles/101000434623-how-can-i-get-an-exact-manufacturer-match-with-specific-suppliers
# Suggests to use:
# options: {filters: {distributor_id: ["459", "1106", "2401", "2628", "2454", "3261", "12947"]}}
# But this isn't a "hard" filter, according to the support people you must filter it anyways.

# Author information.
__author__ = 'Salvador Eduardo Tropea'
__webpage__ = 'https://github.com/set-soft'
__company__ = 'Instituto Nacional de Tecnologia Industrial - Argentina'

# Libraries.
import copy
import difflib
import json
import os
import pprint
import re
import requests
import sys
import time
from collections import OrderedDict
if sys.version_info[0] < 3:
    from urllib import quote_plus
else:
    from urllib.parse import quote_plus

# KiCost definitions.
from .. import KiCostError, DistData, DEFAULT_CURRENCY, ERR_SCRAPE, W_NOINFO, NO_PRICE, W_APIFAIL, W_AMBIPN
# Distributors definitions.
from .distributor import distributor_class, QueryCache, hide_secrets
from .log__ import debug_overview, debug_obsessive, warning


# Uncomment for debug
# Use `debug('x + 1')` for instance.
# def debug(expression):
#     frame = sys._getframe(1)
#     distributor_class.logger.info(expression, '=', repr(eval(expression, frame.f_globals, frame.f_locals)))

MAX_PARTS_PER_QUERY = 20  # Maximum number of parts in a single query.

# Information to return from Nexar API
QUERY_ANSWER = '''
    hits,
    reference,
    parts {
      id,
      slug,
      mpn,
      manufacturer {name, id},
      shortDescription,
      specs {
        attribute {shortname, name, id},
        displayValue
      },
      octopartUrl,
      bestDatasheet {name, url},
      sellers(authorizedOnly: false) {
        company {name, id},
        offers {
            sku,
            id,
            inventoryLevel,
            moq,
            orderMultiple,
            packaging,
            prices {quantity, price, currency},
            onOrderQuantity,
            clickUrl
        }
      }
    }
'''
QUERY_ANSWER = re.sub(r'[\s\n]', '', QUERY_ANSWER)


QUERY_MATCH = ('query MultiMatchSearch($queries: [SupPartMatchQuery!]!) {'
               'supMultiMatch(queries: $queries, country: "@COUNTRY@", currency: "@CUR@") {' +
               QUERY_ANSWER + '} }')
QUERY_URL = 'https://api.nexar.com/graphql'
PROD_TOKEN_URL = 'https://identity.nexar.com/connect/token'
# Specs known by KiCost
SPEC_NAMES = {'tolerance': 'tolerance',
              'frequency': 'frequency',
              'powerrating': 'power',
              'voltagerating': 'voltage',
              'temperaturecoefficient': 'temp_coeff',
              'case_package': 'footprint',
              'maxdccurrent': 'current',
              'forwardcurrent': 'current',
              'currentrating': 'current'}


__all__ = ['api_nexar']


class QueryStruct(object):
    def __init__(self, id, part, for_dists, seller=None, sku=None, manf=None, mpn=None):
        self.id = 'q'+str(id)
        query = OrderedDict()
        query["reference"] = self.id
        query["start"] = 0
        query["limit"] = 5
        self.seller = seller
        self.sku = sku
        self.manf = manf
        self.mpn = mpn
        if sku is not None:
            query['seller'] = seller
            query['sku'] = sku
        elif mpn is not None:
            if manf:
                query['manufacturer'] = manf
            query['mpn'] = mpn
        self.query = query  # The query dict
        self.part = part    # Pointer to the part.
        self.for_dists = for_dists  # List of distributors we want for this query
        self.result = None
        self.loaded = False

    def remove_manufacturer(self):
        self.manf = None
        del self.query['manufacturer']


class api_nexar(distributor_class):
    name = 'Nexar'
    type = 'api'
    enabled = True
    url = 'https://nexar.com/api'  # Web site API information.
    id = None
    secret = None
    country = 'US'
    env_prefix = 'NEXAR'
    env_ops = {'NEXAR_STORAGE_PATH': 'cache_path'}

    # This is what we used for Octopart
    # https://octopart.com/api/v4/values#sellers
    api_distributors = ['arrow', 'digikey', 'farnell', 'lcsc', 'mouser', 'newark', 'rs', 'tme']
    DIST_TRANSLATION = {  # Distributor translation. Just a few supported.
                        'Arrow Electronics': 'arrow',
                        'Digi-Key': 'digikey',
                        'Farnell': 'farnell',
                        'Mouser': 'mouser',
                        'Newark': 'newark',
                        'RS Components': 'rs',
                        'TME': 'tme',
                        'LCSC': 'lcsc',
                       }
    # Dict to translate KiCost field names into KitSpace distributor names
    KICOST2NEXAR_DIST = {v: k for k, v in DIST_TRANSLATION.items()}
    cache = None
    config_options = {'client_id': str, 'client_secret': str, 'country': str}
    # Token
    expiration = 0
    token = None
    access_token = None

    @staticmethod
    def configure(ops):
        cache_ttl = 7
        cache_path = None
        for k, v in ops.items():
            if k == 'enable':
                api_nexar.enabled = v
            elif k == 'cache_ttl':
                cache_ttl = v
            elif k == 'cache_path':
                cache_path = v
            elif k == 'client_id':
                api_nexar.id = v
            elif k == 'client_secret':
                api_nexar.secret = v
            elif k == 'country':
                api_nexar.country = v
        if api_nexar.enabled and (api_nexar.id is None or api_nexar.secret is None or cache_path is None):
            warning(W_APIFAIL, "Can't enable Nexar without a `client_id`, `client_secret` and `cache_path`")
            api_nexar.enabled = False
        debug_obsessive('Nexar API configured to enabled {} id {} secret {} path {}'.
                        format(api_nexar.enabled, hide_secrets(api_nexar.id), hide_secrets(api_nexar.secret), cache_path))
        if not api_nexar.enabled:
            return
        api_nexar.cache = QueryCache(cache_path, cache_ttl)

    @staticmethod
    def get_token():
        """ Return the Nexar token from the client_id and client_secret provided """
        token = {}
        try:
            token = requests.post(url=PROD_TOKEN_URL,
                                  data={"grant_type": "client_credentials",
                                        "client_id": api_nexar.id,
                                        "client_secret": api_nexar.secret},
                                  allow_redirects=False).json()
        except Exception as e:
            raise KiCostError('Error getting token from Nexar ({})'.format(e), ERR_SCRAPE)
        debug_obsessive('Nexar token {}'.format(token))
        return token

    @staticmethod
    def get_headers():
        """ Get the access token, make sure it isn't expired """
        if os.environ.get('KICOST_NEXAR_NO_TOKEN', None):
            # No token for debug
            return {}
        if api_nexar.expiration < time.time() + 300:
            api_nexar.token = api_nexar.get_token()
            api_nexar.access_token = api_nexar.token.get('access_token')
            api_nexar.expiration = time.time() + api_nexar.token.get('expires_in', 0)
        return {'token': api_nexar.access_token}

    @staticmethod
    def query(query_parts, currency, query_type=QUERY_MATCH):
        """ Send query to server and return results. """
        # Allow changing the URL for debug purposes
        try:
            url = os.environ['KICOST_NEXAR_URL']
        except KeyError:
            url = QUERY_URL
        query_type = query_type.replace('@COUNTRY@', api_nexar.country)
        query_type = query_type.replace('@CUR@', currency)
        variables = {"queries": query_parts}
        # Remove all spaces, even inside the manf#
        # SET comment: this is how the code always worked. Octopart (used by KitSpace) ignores spaces inside manf# codes.
        # Do the query using POST
        data = 'query={}&variables={}'.format(quote_plus(query_type), quote_plus(str(variables)))
        distributor_class.log_request(url, data)
        data = OrderedDict()
        data["query"] = query_type
        data["variables"] = variables
        response = requests.post(url, json=data, headers=api_nexar.get_headers())
        response.encoding = 'UTF-8'
        distributor_class.log_response(response)
        if response.status_code == requests.codes['ok']:  # 200
            results = json.loads(response.text)
            return results
        elif response.status_code == requests.codes['not_found']:  # 404
            raise KiCostError('Nexar server not found check your internet connection.', ERR_SCRAPE)
        elif response.status_code == requests.codes['request_timeout']:  # 408
            raise KiCostError('Nexar is not responding.', ERR_SCRAPE)
        elif response.status_code == requests.codes['bad_request']:  # 400
            raise KiCostError('Bad request to Nexar server probably due to an incorrect string '
                              'format check your `manf#` codes and contact the suport team.', ERR_SCRAPE)
        elif response.status_code == requests.codes['gateway_timeout']:  # 504
            raise KiCostError('One of the internal Nexar services may experiencing problems.', ERR_SCRAPE)
        else:
            raise KiCostError('Nexar error: ' + str(response.status_code), ERR_SCRAPE)

    @staticmethod
    def get_spec(data, item, default=None):
        '''Get the value of `value` field of a dictionary if the `name` field identifier.
        Used to get information from the JSON response.'''
        for d in data['specs']:
            if d['attribute']['shortname'] == item:
                return d.get('displayValue', default)
        return default

    @staticmethod
    def query2name(q):
        ''' Finds the prefix and name for a query '''
        if q.mpn is not None:
            prefix = 'mpn'
            name = (q.manf if q.manf else 'unk') + '_' + q.mpn
        elif q.sku is not None:
            prefix = 'sku'
            name = (q.seller if q.seller else 'unk') + '_' + q.sku
        name = name.replace(' ', '_')
        return prefix, name

    @staticmethod
    def get_part_info(queries, currency, to_retry):
        '''Query PartInfo for quantity/price info.
           `distributors` is the list of all distributors we want, in general. '''
        only_query = [q.query for q in queries]
        results = api_nexar.query(only_query, currency)
        for i, r in enumerate(results['data']['supMultiMatch']):
            q = queries[i]
            q.result = r
            assert r['reference'] == q.id, 'Out of order results, please report'
            # Solve the prefix and name
            prefix, name = api_nexar.query2name(q)
            api_nexar.cache.save_results(prefix, name, r)
            if not r.get('hits') and q.mpn is not None and q.manf:
                # Found, but has no hits
                # Try without specifying a manufacturer
                q.remove_manufacturer()
                to_retry.append(q)

    @staticmethod
    def fill_extra_info(result, specs, extra_info):
        # We fill only missing information, the mandatory info comes from the original distributor
        if 'desc' not in extra_info:
            desc = result.get('shortDescription')
            if desc:
                extra_info['desc'] = desc
        if 'manf' not in extra_info:
            manf = result.get('manufacturer', {}).get('name')
            if manf:
                extra_info['manf'] = manf
        if 'value' not in extra_info:
            value = ''
            for spec in ('capacitance', 'resistance', 'inductance'):
                val = specs.get(spec, None)
                if val:
                    value += val[1] + ' '
            if value:
                extra_info['value'] = value
        if 'size' not in extra_info:
            size = ''
            le = specs.get('length')
            if le:
                size += 'L: '+le[1]+' '
            w = specs.get('width')
            if w:
                size += 'W: '+w[1]+' '
            h = specs.get('height')
            if h:
                size += 'H: '+h[1]
            if size:
                extra_info['size'] = size
        for spec, name in SPEC_NAMES.items():
            if name in extra_info:
                continue
            val = specs.get(spec, None)
            if val:
                extra_info[name] = val[1]

    @staticmethod
    def select_best_part(result, part, native_dists):
        """ Select the best part, we discard results that are not distributed by our distributors.
            Then we look for the one that best matches the `manf`. """
        hits = result['hits']
        debug_obsessive('Hits: {}'.format(hits))
        # Each result can match one or more components from different manufacturers
        # Take only the items with useful offers
        useful_items = []
        for item in result['parts']:
            for offer in item['sellers']:
                if offer['company']['name'] in native_dists:
                    useful_items.append(item)
                    break
        # debug_obsessive(str(result))
        # debug_obsessive(str(useful_items))
        # If more than one select the best
        no_offers = False
        if len(useful_items) > 1:
            # List of possible manufacturers
            manufacturers = {}
            for it in useful_items:
                manf = it['manufacturer']['name'].lower()
                if manf not in manufacturers:
                    # We store the only the first one
                    # Nexar provides the best entry first, and sometimes an optional entry
                    # Example: Vishay CRCW06030000Z0EA vs Vishay CRCW0603-0000Z0EA
                    manufacturers[manf] = it
            # Is the manf included?
            manf = part.fields.get('manf', 'none').lower()
            item = manufacturers.get(manf)
            if not item:
                if manf == 'none':
                    item = useful_items[0]
                else:
                    best_matches = difflib.get_close_matches(manf, manufacturers.keys())
                    if best_matches:
                        item = manufacturers[best_matches[0]]
                    else:
                        item = useful_items[0]
                mpn = item['mpn']
                warning(W_AMBIPN, 'Using "{}" for manf#="{}"'.format(item['manufacturer']['name'], mpn))
                warning(W_AMBIPN, 'Ambiguous manf#="{}" please use manf to select the right one, choices: {}'.format(
                        mpn, list(manufacturers.keys())))
        else:
            if len(useful_items):
                item = useful_items[0]
            else:
                item = result['parts'][0]
                no_offers = True
        return item, no_offers

    @staticmethod
    def fill_part_data(part, result):
        ''' Fill generic data in the PartGroup() structure '''
        # Get the information of the part.
        # Datasheet, only if we don't have one
        if not part.datasheet:
            best_datasheet = result.get('bestDatasheet')
            if best_datasheet:
                part.datasheet = best_datasheet.get('url')
        # Life cycle
        lifecycle = api_nexar.get_spec(result, 'lifecyclestatus', '')
        if not lifecycle:
            lifecycle = api_nexar.get_spec(result, 'manufacturerlifecyclestatus', '')
        if lifecycle:
            # End Of Live -> obsolete
            lifecycle = lifecycle.replace('EOL ', 'obsolete ')
            part.lifecycle = lifecycle
        # Misc data collected, currently not used inside KiCost
        part.update_specs({sp['attribute']['shortname']: (sp['attribute']['name'], sp['displayValue'])
                          for sp in result['specs'] if sp['displayValue']})

    @staticmethod
    def fill_part_info(queries, distributors, currency, solved):
        ''' Place the results into the parts list. '''
        # Translate from PartInfo distributor names to the names used internally by kicost.
        dist_xlate = api_nexar.DIST_TRANSLATION

        # Loop through the response to the query and enter info into the parts list.
        for q in queries:
            # Unpack the structure
            part_query = q.query
            part = q.part
            dist_want = q.for_dists
            result = q.result
            # Process it
            if not result or not result.get('hits'):
                warning(W_NOINFO, 'No information found for parts \'{}\' query `{}`'.format(part.refs, str(part_query)))
                continue
            # Select the best hit
            native_dists = set((api_nexar.KICOST2NEXAR_DIST[d] for d in dist_want))
            result, no_offers = api_nexar.select_best_part(result, part, native_dists)
            # Note that it could have no useful offers, in this case we extract some info and then skip it
            api_nexar.fill_part_data(part, result)
            if no_offers:
                continue
            # Loop through the offers from various dists for this particular part.
            for offer in result['sellers']:
                # Get the distributor who made the offer and add their
                # price/qty info to the parts list if its one of the accepted distributors.
                dist = dist_xlate.get(offer['company']['name'], '')
                if dist not in dist_want:
                    # Not interested in this distributor
                    # debug_obsessive('Discard offer: {}'.format(offer['company']['name']))
                    continue
                # Get the DistData for this distributor
                dd = part.dd.get(dist, DistData())
                # Extra information
                api_nexar.fill_extra_info(result, part.specs, dd.extra_info)
                # Each distributor (seller) can have one or more offers
                for of in offer['offers']:
                    # Get pricing information from this distributor.
                    # The offer could contain more than one currency, so we separate the prices by currency.
                    dist_currency = {}
                    for pr in of['prices']:
                        cur = pr['currency']
                        ne = (pr['quantity'], pr['price'])
                        if cur in dist_currency:
                            dist_currency[cur].append(ne)
                        else:
                            dist_currency[cur] = [ne]
                    if not dist_currency:
                        # Some times the API returns minimum purchase 0 and a not valid `price_tiers`.
                        warning(NO_PRICE, 'No price information found for parts \'{}\' query `{}`'.format(part.refs, str(part_query)))
                    else:
                        prices = None
                        # Get the price tiers prioritizing:
                        # 1) The asked currency by KiCost user;
                        # 2) The default currency given by `DEFAULT_CURRENCY` in root `global_vars.py`;
                        # 3) The first not null tiers
                        if currency in dist_currency:
                            prices = dist_currency[currency]
                            dd.currency = currency
                        elif DEFAULT_CURRENCY in dist_currency:
                            prices = dist_currency[DEFAULT_CURRENCY]
                            dd.currency = DEFAULT_CURRENCY
                        else:
                            dd.currency, prices = next(iter(dist_currency.items()))
                        price_tiers = {qty: float(price) for qty, price in prices}
                        # Combine price lists for multiple offers from the same distributor
                        # to build a complete list of cut-tape and reeled components.
                        dd.price_tiers.update(price_tiers)
                    part_qty_increment = of.get('orderMultiple', 1)
                    # Select the part SKU, web page, and available quantity.
                    # Each distributor can have different stock codes for the same part in different
                    # quantities / delivery package styles: cut-tape, reel, ...
                    # Therefore we select and overwrite a previous selection if one of the
                    # following conditions is met:
                    #   1. We don't have a selection for this part from this distributor yet.
                    #   2. The MOQ is smaller than for the current selection.
                    #   3. The part_qty_increment for this offer smaller than that of the existing selection.
                    #      (we prefer cut-tape style packaging over reels)
                    #   4. For DigiKey, we can't use part_qty_increment to distinguish between
                    #      reel and cut-tape, so we need to look at the actual DigiKey part number.
                    #      This procedure is made by the definition `distributors_info[dist]['ignore_cat#_re']`
                    #      at the distributor profile.
                    dist_part_num = of.get('sku', '')
                    qty_avail = dd.qty_avail
                    in_stock_quantity = of.get('inventoryLevel')
                    if not qty_avail or (in_stock_quantity and qty_avail < in_stock_quantity):
                        # Keeps the information of more availability.
                        dd.qty_avail = in_stock_quantity  # In stock.
                    ign_stock_code = distributor_class.get_distributor_info(dist).ignore_cat
                    valid_part = not (ign_stock_code and re.match(ign_stock_code, dist_part_num))
                    # debug('dd.part_num')  # Uncomment to debug
                    # debug('dd.qty_increment')  # Uncomment to debug
                    moq = of.get('moq')
                    if (valid_part and
                        (not dd.part_num or
                         (dd.qty_increment is None or part_qty_increment < dd.qty_increment) or
                         (not dd.moq or (moq and dd.moq > moq)))):
                        # Save the link, stock code, ... of the page for minimum purchase.
                        dd.moq = moq  # Minimum order qty.
                        dd.url = of.get('clickUrl', '')  # Page to purchase the minimum quantity.
                        dd.part_num = dist_part_num
                        dd.qty_increment = part_qty_increment
                    # Update the DistData for this distributor
                    part.dd[dist] = dd
                # We have data for this distributor
                solved.add(dist)

    @staticmethod
    def look_query_in_cache(query, unsolved):
        # Solve the prefix and name
        prefix, name = api_nexar.query2name(query)
        # Look in the cache
        query.result, query.loaded = api_nexar.cache.load_results(prefix, name)
        if not query.loaded:
            unsolved.append(query)
        else:
            debug_obsessive('Data from cache: '+pprint.pformat(query.result))

    @staticmethod
    def solve_queries(unsolved, n_unsolved, currency):
        to_retry = []
        # Setup progress bar to track progress of server queries.
        progress = distributor_class.progress(n_unsolved, distributor_class.logger)

        # Slice the queries into batches of the largest allowed size and gather
        # the part data for each batch.
        for i in range(0, n_unsolved, MAX_PARTS_PER_QUERY):
            slc = slice(i, i+MAX_PARTS_PER_QUERY)
            api_nexar.get_part_info(unsolved[slc], currency, to_retry)
            progress.update(len(unsolved[slc]))

        # Done with the scraping progress bar so delete it or else we get an
        # error when the program terminates.
        progress.close()
        return to_retry, len(to_retry)

    @staticmethod
    def query_part_info(parts, distributors, currency):
        '''Fill-in the parts with price/qty/etc info from Nexar.'''
        debug_overview('# Getting part data from Nexar ...')

        # Use just the distributors avaliable in this API.
        # Note: The user can use --exclude and define it with fields.
        distributors = [d for d in distributors if distributor_class.get_distributor_info(d).is_web()
                        and d in api_nexar.api_distributors]
        FIELDS_CAT = sorted([d + '#' for d in distributors])

        # Create queries to get part price/quantities from PartInfo.
        queries = []
        # Translate from PartInfo distributor names to the names used internally by kicost.
        available_distributors = set(api_nexar.api_distributors)
        for part in parts:
            # Create a PartInfo query using the manufacturer's part number or the distributor's SKU.
            part_dist_use_manfpn = copy.copy(distributors)

            # Create queries using the distributor SKU
            # Check if that part have stock code that is accepted to use by this module (API).
            # KiCost will prioritize these codes under "manf#" that will be used for get
            # information for the part hat were not filled with the distributor stock code. So
            # this is checked after the 'manf#' buv code.
            found_codes_for_all_dists = True
            for d in FIELDS_CAT:
                part_stock = part.fields.get(d)
                if part_stock:
                    part_catalogue_code_dist = d[:-1]
                    if part_catalogue_code_dist in available_distributors:
                        part_code_dist = api_nexar.KICOST2NEXAR_DIST[part_catalogue_code_dist]
                        queries.append(QueryStruct(len(queries), part, [part_catalogue_code_dist], seller=part_code_dist, sku=part_stock))
                        part_dist_use_manfpn.remove(part_catalogue_code_dist)
                else:
                    found_codes_for_all_dists = False

            # Create a query using the manufacturer P/N
            part_manf = part.fields.get('manf', '')
            part_code = part.fields.get('manf#')
            if part_code and not found_codes_for_all_dists:
                # Not all distributors has code, add a query for the manufaturer P/N
                # List of distributors without an specific part number
                queries.append(QueryStruct(len(queries), part, part_dist_use_manfpn, manf=part_manf, mpn=part_code))

        n_queries = len(queries)
        debug_overview('Queries {}'.format(n_queries))
        if not n_queries:
            return

        # Try to solve the queries from the cache
        unsolved = []
        for query in queries:
            api_nexar.look_query_in_cache(query, unsolved)
            if query.loaded and not query.result.get('hits') and query.mpn is not None and query.manf:
                # Found, but has no hits
                # Try without specifying a manufacturer
                query.remove_manufacturer()
                api_nexar.look_query_in_cache(query, unsolved)

        # Solve the rest from the site
        n_unsolved = len(unsolved)
        debug_overview('Cached entries {} (of {})'.format(n_queries-n_unsolved, n_queries))
        if n_unsolved:
            unsolved, n_unsolved = api_nexar.solve_queries(unsolved, n_unsolved, currency)
            # Do we need to retry some queries?
            if n_unsolved:
                unsolved, n_unsolved = api_nexar.solve_queries(unsolved, n_unsolved, currency)

        # Transfer the results
        solved_dist = set()
        api_nexar.fill_part_info(queries, distributors, currency, solved_dist)
        return solved_dist


distributor_class.register(api_nexar, 75)
